from netCDF4 import Dataset
import numpy as np
import pandas as pd
import json
from glob import glob
import os
from datetime import datetime
from multiprocessing import Pool
import argparse
import pickle
from scipy.ndimage import binary_dilation
from scipy.stats import multivariate_normal
import traceback


def load_grid_info(grid_file):
    grid_info = {}
    grid_obj = Dataset(grid_file)
    for var in grid_obj.variables.keys():
        grid_info[var] = grid_obj.variables[var][:]
    grid_obj.close()
    return grid_info


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("run_dates", help="Model Run dates YYYYMMDD")
    parser.add_argument("-p", "--proc", type=int, help="Number of processors")
    args = parser.parse_args()
    run_dates_str = args.run_dates.split(",")
    run_dates = [datetime.strptime(rd, "%Y%m%d") for rd in run_dates_str]
    members = ["wrf-s3cn_arw"] + ["wrf-s3m{0:d}_arw".format(m) for m in range(3, 14)]
    model_names = ["Random-Forest", "Gradient-Boosting", "Random-Forest-Weighted", "Random-Forest-Big"]
    start_hour = 12
    end_hour = 36
    dx = 3000
    num_samples = 1000
    thresholds = np.array([0, 25, 50])
    track_path = "/sharp/djgagne/track_forecasts_spring2015_json/"
    out_path = "/sharp/djgagne/track_samples_spring2015/"
    grid_filename = "/sharp/djgagne/ssef_2015_grid.nc"
    copula_file = "/sharp/djgagne/track_copulas.pkl"
    member_info_file = "/sharp/djgagne/member_info_ssef_spring2015.csv"
    member_info = pd.read_csv(member_info_file, index_col="Ensemble_Member")
    grid_info = load_grid_info(grid_filename)
    grid_shape = grid_info["lon"].shape
    if args.proc > 1:
        pool = Pool(args.proc)
        try:
            for run_date in run_dates:
                for member in members:
                    group = member_info.loc[member, "Microphysics"]
                    pool.apply_async(sample_member_run_tracks, (member,
                                                                group,
                                                                run_date,
                                                                model_names,
                                                                start_hour,
                                                                end_hour,
                                                                grid_shape,
                                                                dx,
                                                                track_path,
                                                                num_samples,
                                                                thresholds,
                                                                copula_file,
                                                                out_path
                                                                ))
            pool.close()
            pool.join()
        finally:
            pool.terminate()
            pool.join()
    else:
        for run_date in run_dates:
            for member in members:
                group = member_info.loc[member, "Microphysics"]
                apply(sample_member_run_tracks, (member,
                                                 group,
                                                 run_date,
                                                 model_names,
                                                 start_hour,
                                                 end_hour,
                                                 grid_shape,
                                                 dx,
                                                 track_path,
                                                 num_samples,
                                                 thresholds,
                                                 copula_file,
                                                 out_path
                                                 ))

    return


def sample_member_run_tracks(member,
                             group,
                             run_date,
                             model_names,
                             start_hour,
                             end_hour,
                             grid_shape,
                             dx,
                             track_path,
                             num_samples,
                             thresholds,
                             copula_file,
                             out_path,
                             size_ranges,
                             track_ranges):
    try:
        ts = TrackSampler(member,
                          group,
                          run_date,
                          model_names,
                          start_hour,
                          end_hour,
                          grid_shape,
                          dx,
                          track_path,
                          num_samples,
                          copula_file=copula_file
                          )
        ts.load_track_forecasts()
        track_probs = ts.sample_tracks(size_ranges, track_ranges, thresholds)
        ts.output_track_probs(track_probs, out_path)
    except Exception as e:
        print(traceback.format_exc())
        raise e
    return


class TrackSampler(object):
    """
    Monte Carlo sampler of forecast storm tracks.

    """
    def __init__(self, member, group, run_date, model_names, start_hour,
                 end_hour, grid_shape, dx, track_path, num_samples, copula_file=None):
        self.member = member
        self.group = group
        self.run_date = run_date
        self.model_names = [mn.replace(" ", "-") for mn in model_names]
        self.track_path = track_path
        self.num_samples = num_samples
        self.start_hour = start_hour
        self.end_hour = end_hour
        self.grid_shape = grid_shape
        self.dx = int(dx)
        self.copula_file = copula_file
        if self.copula_file is not None:
            with open(self.copula_file) as copula_obj:
                self.copula = pickle.load(copula_obj)
        self.track_forecasts = []

    def load_track_forecasts(self):
        run_date_str = self.run_date.strftime("%Y%m%d")
        track_files = sorted(glob(self.track_path + "/".join([run_date_str, self.member]) + "/*.json"))
        for track_file in track_files:
            tfo = open(track_file)
            self.track_forecasts.append(json.load(tfo))
            tfo.close()

    def generate_copula_ranks(self):
        copula_samples = multivariate_normal.rvs(self.copula[self.group]['mean'],
                                                 self.copula[self.group]['cov'],
                                                 self.num_samples)
        ranks = np.argsort(copula_samples, axis=0)
        return pd.DataFrame(data=ranks, columns=self.copula[self.group]['model_names'])

    def sample_tracks(self, size_ranges, track_ranges, thresholds=np.array([0, 25, 50]), dilation=13):
        track_probs = {}
        print(self.member, "Sample condition")
        condition_samples = self.sample_condition()
        print(self.member, "Sample size")
        size_values = np.arange(size_ranges[0], size_ranges[1] + size_ranges[2], size_ranges[2], dtype=int)
        size_samples = self.sample_size(size_values=size_values)
        print(self.member, "Sample start time")
        start_time_values = np.arange(track_ranges['start-time'][0],
                                      track_ranges['start-time'][1] + track_ranges['start-time'][2],
                                      track_ranges['start-time'][2], dtype=int)
        st_samples = self.sample_start_time(start_time_values=start_time_values)
        print(self.member, "Sample translation")
        translation_x_values = np.arange(track_ranges['translation-x'][0],
                                         track_ranges['translation-x'][1] + track_ranges['translation-x'][2],
                                         track_ranges['translation-x'][2], dtype=int)
        translation_y_values = np.arange(track_ranges['translation-y'][0],
                                         track_ranges['translation-y'][1] + track_ranges['translation-y'][2],
                                         track_ranges['translation-y'][2], dtype=int)
        tx_samples = self.sample_translation_x(translation_x_values=translation_x_values)
        ty_samples = self.sample_translation_y(translation_y_values=translation_y_values)
        if self.copula_file is not None:
            copula_ranks = self.generate_copula_ranks()
        for model_name in self.model_names:
            print(model_name)
            track_probs[model_name] = {}
            for thresh in thresholds:
                track_probs[model_name][thresh] = np.zeros((self.end_hour - self.start_hour + 1,
                                                            self.grid_shape[0],
                                                            self.grid_shape[1]))
            for t, track in enumerate(self.track_forecasts):
                print(model_name, self.member, self.run_date, t)
                start_time = track['properties']['times'][0]
                st_track_sample = np.zeros(st_samples[model_name][t].shape, dtype=int)
                tx_track_sample = np.zeros(tx_samples[model_name][t].shape, dtype=int)
                ty_track_sample = np.zeros(ty_samples[model_name][t].shape, dtype=int)
                if self.copula_file is not None:
                    print("copula being used")
                    st_track_sample[copula_ranks['start-time'].values] = np.sort(st_samples[model_name][t])
                    tx_track_sample[copula_ranks['translation-x'].values] = np.sort(tx_samples[model_name][t])
                    ty_track_sample[copula_ranks['translation-y'].values] = np.sort(ty_samples[model_name][t])
                else:
                    print("copula not being used")
                    st_track_sample = st_samples[model_name][t]
                    tx_track_sample = st_samples[model_name][t]
                    ty_track_sample = st_samples[model_name][t]
                for f, feature in enumerate(track['features']):
                    i_track = np.array(feature['properties']['i'], dtype=int)
                    j_track = np.array(feature['properties']['j'], dtype=int)
                    j_track_big, i_track_big = np.meshgrid(
                        np.arange(j_track.min() - dilation, j_track.max() + dilation + 1),
                        np.arange(i_track.min() - dilation, i_track.max() + dilation + 1))
                    mask = np.array(feature['properties']['masks'], dtype=int)
                    mask_big = np.zeros((mask.shape[0] + dilation * 2, mask.shape[1] + dilation * 2))
                    mask_big[dilation: (dilation + mask.shape[0]), dilation: (dilation + mask.shape[1])] = mask
                    mask_big[binary_dilation(mask_big, iterations=dilation)] = 1
                    i_good = i_track_big[mask_big > 0]
                    j_good = j_track_big[mask_big > 0]
                    for s in range(self.num_samples):
                        st_corr = start_time + st_track_sample[s]
                        if (condition_samples[model_name][t][f][s] > 0) and (st_corr >= self.start_hour) \
                                and (st_corr <= self.end_hour):
                            tj_corr = tx_track_sample[s] // self.dx
                            ti_corr = ty_track_sample[s] // self.dx
                            size_sample = size_samples[model_name][t][f][s]
                            i_adj = i_good + ti_corr
                            j_adj = j_good + tj_corr
                            ij_valid = (i_adj < self.grid_shape[0]) & (i_adj >= 0) & \
                                       (j_adj < self.grid_shape[1]) & (j_adj >= 0)
                            i_adj = i_adj[ij_valid]
                            j_adj = j_adj[ij_valid]
                            for thresh in thresholds:
                                if size_sample >= thresh and ((st_corr + f) <= self.end_hour):
                                    track_probs[model_name][thresh][st_corr + f - self.start_hour,
                                                                    i_adj, j_adj] += 1
            for thresh in thresholds:
                track_probs[model_name][thresh] /= float(self.num_samples)
        return track_probs

    def sample_condition(self):
        condition_samples = {}
        for model_name in self.model_names:
            condition_samples[model_name] = []
            for t, track in enumerate(self.track_forecasts):
                condition_samples[model_name].append([])
                for f, feature in enumerate(track['features']):
                    cond_prob = feature['properties']['condition_' + model_name]
                    condition_samples[model_name][t].append(np.random.choice([0, 1],
                                                                             self.num_samples,
                                                                             p=[1 - cond_prob, cond_prob]))
        return condition_samples

    def sample_size(self, size_values=np.arange(5, 105, 5)):
        size_samples = {}
        for model_name in self.model_names:
            size_samples[model_name] = []
            for t, track in enumerate(self.track_forecasts):
                size_samples[model_name].append([])
                for f, feature in enumerate(track['features']):
                    size_prob = feature['properties']['size_' + model_name]
                    size_samples[model_name][t].append(np.random.choice(size_values,
                                                                        self.num_samples,
                                                                        p=size_prob))
        return size_samples

    def sample_start_time(self, start_time_values=np.arange(-6, 7, 1)):
        st_samples = {}
        for model_name in self.model_names:
            st_samples[model_name] = []
            for t, track in enumerate(self.track_forecasts):
                st_prob = track['features'][0]['properties']['start-time_' + model_name]
                st_samples[model_name].append(np.random.choice(start_time_values,
                                                               self.num_samples,
                                                               p=st_prob))
        return st_samples

    def sample_translation_x(self, translation_x_values=np.arange(-192000, 192000 + 24000, 24000)):
        tx_samples = {}
        for model_name in self.model_names:
            tx_samples[model_name] = []
            for t, track in enumerate(self.track_forecasts):
                tx_prob = track['features'][0]['properties']['translation-x_' + model_name]
                tx_samples[model_name].append(np.random.choice(translation_x_values,
                                                               self.num_samples,
                                                               p=tx_prob) +
                                              np.random.randint(0,
                                                                translation_x_values[1] - translation_x_values[0],
                                                                self.num_samples))
        return tx_samples

    def sample_translation_y(self, translation_y_values=np.arange(-192000, 192000 + 24000, 24000)):
        ty_samples = {}
        for model_name in self.model_names:
            ty_samples[model_name] = []
            for t, track in enumerate(self.track_forecasts):
                ty_prob = track['features'][0]['properties']['translation-y_' + model_name]
                ty_samples[model_name].append(np.random.choice(translation_y_values,
                                                               self.num_samples,
                                                               p=ty_prob) +
                                              np.random.randint(0,
                                                                translation_y_values[1] - translation_y_values[0],
                                                                self.num_samples))
        return ty_samples

    def output_track_probs(self, track_probs, path):
        run_date_str = self.run_date.strftime("%Y%m%d")
        if not os.access(path + run_date_str, os.R_OK):
            try:
                os.mkdir(path + run_date_str)
            except OSError:
                print(path + run_date_str + " already created")
        for model in self.model_names:
            filename = path + "{2}/{0}_hailprobs_{1}_{2}.nc".format(model, self.member, run_date_str)
            out_obj = Dataset(filename, "w")
            out_obj.createDimension("hour", self.end_hour - self.start_hour + 1)
            out_obj.createDimension("y", self.grid_shape[0])
            out_obj.createDimension("x", self.grid_shape[1])
            forecast_hour = out_obj.createVariable("forecast_hour", "i4", ("hour",))
            forecast_hour[:] = np.arange(self.start_hour, self.end_hour + 1)
            forecast_hour.units = "hour"
            for thresh in track_probs[model].keys():
                var_obj = out_obj.createVariable("prob_hail_{0:02d}_mm".format(thresh),
                                                 "f4",
                                                 ("hour", "y", "x"), zlib=True)
                var_obj[:] = track_probs[model][thresh]
                var_obj.units = ""
                var_obj.long_name = "probability of hail at least {0:02d} mm diameter".format(thresh)
            out_obj.model_name = model
            out_obj.run_date = run_date_str
            out_obj.close()
        return


if __name__ == "__main__":
    main()



import numpy as np
from hagelslag.util.munkres import Munkres
import pandas as pd


class ObjectMatcher(object):
    """
    ObjectMatcher calculates distances between two sets of objects and determines the optimal object assignments
    based on the Hungarian object matching algorithm. ObjectMatcher supports the use of the weighted average of
    multiple cost functions to determine the distance between objects. Upper limits to each distance component are used
    to exclude the matching of objects that are too far apart.

    Attributes:
        cost_function_components: List of distance functions for matching
        weights: List of weights for each distance function
        max_values : List of the maximum allowable distance for each distance function component.
    """

    def __init__(self, cost_function_components, weights, max_values):
        self.cost_function_components = cost_function_components
        self.weights = weights
        self.max_values = max_values
        if self.weights.sum() != 1:
            self.weights /= float(self.weights.sum())
        return

    def match_objects(self, set_a, set_b, time_a, time_b):
        """
        Match two sets of objects at particular times.

        Args:
            set_a: list of STObjects
            set_b: list of STObjects
            time_a: time at which set_a is being evaluated for matching
            time_b: time at which set_b is being evaluated for matching

        Returns:
            List of tuples containing (set_a index, set_b index) for each match
        """
        costs = self.cost_matrix(set_a, set_b, time_a, time_b) * 100
        min_row_costs = costs.min(axis=1)
        min_col_costs = costs.min(axis=0)
        good_rows = np.where(min_row_costs < 100)[0]
        good_cols = np.where(min_col_costs < 100)[0]
        assignments = []
        if len(good_rows) > 0 and len(good_cols) > 0:
            munk = Munkres()
            initial_assignments = munk.compute(costs[tuple(np.meshgrid(good_rows, good_cols, indexing='ij'))].tolist())
            initial_assignments = [(good_rows[x[0]], good_cols[x[1]]) for x in initial_assignments]
            for a in initial_assignments:
                if costs[a[0], a[1]] < 100:
                    assignments.append(a)
        return assignments

    def cost_matrix(self, set_a, set_b, time_a, time_b):
        """
        Calculates the costs (distances) between the items in set a and set b at the specified times.

        Args:
            set_a: List of STObjects
            set_b: List of STObjects
            time_a: time at which objects in set_a are evaluated
            time_b: time at whcih object in set_b are evaluated

        Returns:
            A numpy array with shape [len(set_a), len(set_b)] containing the cost matrix between the items in set a
            and the items in set b.
        """
        costs = np.zeros((len(set_a), len(set_b)))
        for a, item_a in enumerate(set_a):
            for b, item_b in enumerate(set_b):
                costs[a, b] = self.total_cost_function(item_a, item_b, time_a, time_b)
        return costs

    def total_cost_function(self, item_a, item_b, time_a, time_b):
        """
        Calculate total cost function between two items.

        Args:
            item_a: STObject
            item_b: STObject
            time_a: Timestep in item_a at which cost function is evaluated
            time_b: Timestep in item_b at which cost function is evaluated

        Returns:
            The total weighted distance between item_a and item_b
        """
        distances = np.zeros(len(self.weights))
        for c, component in enumerate(self.cost_function_components):
            distances[c] = component(item_a, time_a, item_b, time_b, self.max_values[c])
        total_distance = np.sum(self.weights * distances)
        return total_distance


class TrackMatcher(object):
    """
    Find the optimal pairings among two sets of STObject tracks.

    Attributes:
        cost_function_components: Array of cost function objects
        weights: Array of weights for each cost function. All should sum to 1.
        max_values: Array of distance values that correspond to the upper limit distance that should be
            considered.

    """

    def __init__(self, cost_function_components, weights, max_values):
        self.cost_function_components = cost_function_components
        self.weights = weights if weights.sum() == 1 else weights / weights.sum()
        self.max_values = max_values

    def match_tracks(self, set_a, set_b, closest_matches=False):
        """
        Find the optimal set of matching assignments between set a and set b. This function supports optimal 1:1
        matching using the Munkres method and matching from every object in set a to the closest object in set b.
        In this situation set b accepts multiple matches from set a.

        Args:
            set_a:
            set_b:
            closest_matches:

        Returns:

        """
        costs = self.track_cost_matrix(set_a, set_b) * 100
        min_row_costs = costs.min(axis=1)
        min_col_costs = costs.min(axis=0)
        good_rows = np.where(min_row_costs < 100)[0]
        good_cols = np.where(min_col_costs < 100)[0]
        assignments = []
        if len(good_rows) > 0 and len(good_cols) > 0:
            if closest_matches:
                b_matches = costs[np.meshgrid(good_rows, good_cols, indexing='ij')].argmin(axis=1)
                a_matches = np.arange(b_matches.size)
                initial_assignments = [(good_rows[a_matches[x]], good_cols[b_matches[x]])
                                       for x in range(b_matches.size)]
            else:
                munk = Munkres()
                initial_assignments = munk.compute(costs[np.meshgrid(good_rows, good_cols, indexing='ij')].tolist())
                initial_assignments = [(good_rows[x[0]], good_cols[x[1]]) for x in initial_assignments]
            for a in initial_assignments:
                if costs[a[0], a[1]] < 100:
                    assignments.append(a)
        return assignments

    def raw_cost_matrix(self, set_a, set_b):
        cost_matrix = np.zeros((len(set_a), len(set_b), len(self.cost_function_components)))
        for (a, b, c), x in np.ndenumerate(cost_matrix):
            cost_matrix[a, b, c] = self.cost_function_components[c](set_a[a],
                                                                    set_b[b], self.max_values[c]) * self.max_values[c]
        return cost_matrix

    def neighbor_matches(self, set_a, set_b):
        costs = self.track_cost_matrix(set_a, set_b)
        all_neighbors = []
        for i in range(len(set_a)):
            neighbors = np.where(costs[i] < 1)[0]
            sorted_neighbors = neighbors[costs[i][neighbors].argsort()]
            if len(neighbors) > 0:
                all_neighbors.append((i, tuple(sorted_neighbors)))
        return all_neighbors

    def track_cost_matrix(self, set_a, set_b):
        costs = np.zeros((len(set_a), len(set_b)))
        for a, item_a in enumerate(set_a):
            for b, item_b in enumerate(set_b):
                costs[a, b] = self.track_cost_function(item_a, item_b)
        return costs

    def track_cost_function(self, item_a, item_b):
        distances = np.zeros(len(self.weights))
        for c, component in enumerate(self.cost_function_components):
            distances[c] = component(item_a, item_b, self.max_values[c])
        if np.all(distances < 1):
            total_distance = np.sum(self.weights * distances)
        else:
            total_distance = 1.0
        return total_distance


class TrackStepMatcher(object):
    """
    Determine if each step in a track is in close proximity to steps from another set of tracks
    """
    def __init__(self, cost_function_components, max_values):
        self.cost_function_components = cost_function_components
        self.max_values = max_values

    def match(self, set_a, set_b):
        """
        For each step in each track from set_a, identify all steps in all tracks from set_b that meet all
        cost function criteria
        
        Args:
            set_a: List of STObjects
            set_b: List of STObjects

        Returns:
            track_pairings: pandas.DataFrame 
        """
        track_step_matches = [[] * len(set_a)]

        costs = self.cost_matrix(set_a, set_b)
        valid_costs = np.all(costs < 1, axis=2)
        set_a_matches, set_b_matches = np.where(valid_costs)
        s = 0
        track_pairings = pd.DataFrame(index=np.arange(costs.shape[0]),
                                      columns=["Track", "Step", "Time", "Matched", "Pairings"], dtype=object)
        set_b_info = []
        for trb, track_b in enumerate(set_b):
            for t, time in enumerate(track_b.times):
                set_b_info.append((trb, t))
        set_b_info_arr = np.array(set_b_info, dtype=int)
        for tr, track_a in enumerate(set_a):
            for t, time in enumerate(track_a.times):
                track_pairings.loc[s, ["Track", "Step", "Time"]] = [tr, t, time]
                track_pairings.loc[s, "Matched"] = 1 if np.count_nonzero(set_a_matches == s) > 0 else 0
                if track_pairings.loc[s, "Matched"] == 1:
                    track_pairings.loc[s, "Pairings"] = set_b_info_arr[set_b_matches[set_a_matches == s]]
                else:
                    track_pairings.loc[s, "Pairings"] = np.array([])
                s += 1
        return track_pairings

    def cost_matrix(self, set_a, set_b):
        num_steps_a = np.sum([track_a.times.size for track_a in set_a])
        num_steps_b = np.sum([track_b.times.size for track_b in set_b])
        cost_matrix = np.zeros((num_steps_a, num_steps_b, len(self.cost_function_components)))
        a_i = 0
        for a, track_a in enumerate(set_a):
            for time_a in track_a.times:
                b_i = 0
                for b, track_b in enumerate(set_b):
                    for time_b in track_b.times:
                        cost_matrix[a_i, b_i] = self.cost(track_a, time_a, track_b, time_b)
                        b_i += 1
                a_i += 1
        return cost_matrix

    def cost(self, track_a, time_a, track_b, time_b):
        return np.array([cost_func(track_a, time_a, track_b, time_b, self.max_values[c])
                         for c, cost_func in enumerate(self.cost_function_components)])


def centroid_distance(item_a, time_a, item_b, time_b, max_value):
    """
    Euclidean distance between the centroids of item_a and item_b.

    Args:
        item_a: STObject from the first set in ObjectMatcher
        time_a: Time integer being evaluated
        item_b: STObject from the second set in ObjectMatcher
        time_b: Time integer being evaluated
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    ax, ay = item_a.center_of_mass(time_a)
    bx, by = item_b.center_of_mass(time_b)
    return np.minimum(np.sqrt((ax - bx) ** 2 + (ay - by) ** 2), max_value) / float(max_value)


def time_distance(item_a, time_a, item_b, time_b, max_value):
    return np.minimum(np.abs(time_b - time_a), max_value) / float(max_value)


def shifted_centroid_distance(item_a, time_a, item_b, time_b, max_value):
    """
    Centroid distance with motion corrections.

    Args:
        item_a: STObject from the first set in ObjectMatcher
        time_a: Time integer being evaluated
        item_b: STObject from the second set in ObjectMatcher
        time_b: Time integer being evaluated
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    ax, ay = item_a.center_of_mass(time_a)
    bx, by = item_b.center_of_mass(time_b)
    if time_a < time_b:
        bx = bx - item_b.u
        by = by - item_b.v
    else:
        ax = ax - item_a.u
        ay = ay - item_a.v
    return np.minimum(np.sqrt((ax - bx) ** 2 + (ay - by) ** 2), max_value) / float(max_value)


def closest_distance(item_a, time_a, item_b, time_b, max_value):
    """
    Euclidean distance between the pixels in item_a and item_b closest to each other.

    Args:
        item_a: STObject from the first set in ObjectMatcher
        time_a: Time integer being evaluated
        item_b: STObject from the second set in ObjectMatcher
        time_b: Time integer being evaluated
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    return np.minimum(item_a.closest_distance(time_a, item_b, time_b), max_value) / float(max_value)


def ellipse_distance(item_a, time_a, item_b, time_b, max_value):
    """
    Calculate differences in the properties of ellipses fitted to each object.

    Args:
        item_a: STObject from the first set in ObjectMatcher
        time_a: Time integer being evaluated
        item_b: STObject from the second set in ObjectMatcher
        time_b: Time integer being evaluated
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    ts = np.array([0, np.pi])
    ell_a = item_a.get_ellipse_model(time_a)
    ell_b = item_b.get_ellipse_model(time_b)
    ends_a = ell_a.predict_xy(ts)
    ends_b = ell_b.predict_xy(ts)
    distances = np.sqrt((ends_a[:, 0:1] - ends_b[:, 0:1].T) ** 2 + (ends_a[:, 1:] - ends_b[:, 1:].T) ** 2)
    return np.minimum(distances[0, 1], max_value) / float(max_value)


def nonoverlap(item_a, time_a, item_b, time_b, max_value):
    """
    Percentage of pixels in each object that do not overlap with the other object

    Args:
        item_a: STObject from the first set in ObjectMatcher
        time_a: Time integer being evaluated
        item_b: STObject from the second set in ObjectMatcher
        time_b: Time integer being evaluated
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    return np.minimum(1 - item_a.count_overlap(time_a, item_b, time_b), max_value) / float(max_value)


def max_intensity(item_a, time_a, item_b, time_b, max_value):
    """
    RMS difference in maximum intensity

    Args:
        item_a: STObject from the first set in ObjectMatcher
        time_a: Time integer being evaluated
        item_b: STObject from the second set in ObjectMatcher
        time_b: Time integer being evaluated
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    intensity_a = item_a.max_intensity(time_a)
    intensity_b = item_b.max_intensity(time_b)
    diff = np.sqrt((intensity_a - intensity_b) ** 2)
    return np.minimum(diff, max_value) / float(max_value)


def area_difference(item_a, time_a, item_b, time_b, max_value):
    """
    RMS Difference in object areas.

    Args:
        item_a: STObject from the first set in ObjectMatcher
        time_a: Time integer being evaluated
        item_b: STObject from the second set in ObjectMatcher
        time_b: Time integer being evaluated
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    size_a = item_a.size(time_a)
    size_b = item_b.size(time_b)
    diff = np.sqrt((size_a - size_b) ** 2)
    return np.minimum(diff, max_value) / float(max_value)


def mean_minimum_centroid_distance(item_a, item_b, max_value):
    """
    RMS difference in the minimum distances from the centroids of one track to the centroids of another track

    Args:
        item_a: STObject from the first set in TrackMatcher
        item_b: STObject from the second set in TrackMatcher
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    centroids_a = np.array([item_a.center_of_mass(t) for t in item_a.times])
    centroids_b = np.array([item_b.center_of_mass(t) for t in item_b.times])
    distance_matrix = (centroids_a[:, 0:1] - centroids_b.T[0:1]) ** 2 + (centroids_a[:, 1:] - centroids_b.T[1:]) ** 2
    mean_min_distances = np.sqrt(distance_matrix.min(axis=0).mean() + distance_matrix.min(axis=1).mean())
    return np.minimum(mean_min_distances, max_value) / float(max_value)


def mean_min_time_distance(item_a, item_b, max_value):
    """
    Calculate the mean time difference among the time steps in each object.

    Args:
        item_a: STObject from the first set in TrackMatcher
        item_b: STObject from the second set in TrackMatcher
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    times_a = item_a.times.reshape((item_a.times.size, 1))
    times_b = item_b.times.reshape((1, item_b.times.size))
    distance_matrix = (times_a - times_b) ** 2
    mean_min_distances = np.sqrt(distance_matrix.min(axis=0).mean() + distance_matrix.min(axis=1).mean())
    return np.minimum(mean_min_distances, max_value) / float(max_value)


def start_centroid_distance(item_a, item_b, max_value):
    """
    Distance between the centroids of the first step in each object.

    Args:
        item_a: STObject from the first set in TrackMatcher
        item_b: STObject from the second set in TrackMatcher
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    start_a = item_a.center_of_mass(item_a.times[0])
    start_b = item_b.center_of_mass(item_b.times[0])
    start_distance = np.sqrt((start_a[0] - start_b[0]) ** 2 + (start_a[1] - start_b[1]) ** 2)
    return np.minimum(start_distance, max_value) / float(max_value)


def start_time_distance(item_a, item_b, max_value):
    """
    Absolute difference between the starting times of each item.

    Args:
        item_a: STObject from the first set in TrackMatcher
        item_b: STObject from the second set in TrackMatcher
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    start_time_diff = np.abs(item_a.times[0] - item_b.times[0])
    return np.minimum(start_time_diff, max_value) / float(max_value)


def duration_distance(item_a, item_b, max_value):
    """
    Absolute difference in the duration of two items

    Args:
        item_a: STObject from the first set in TrackMatcher
        item_b: STObject from the second set in TrackMatcher
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    duration_a = item_a.times.size
    duration_b = item_b.times.size
    return np.minimum(np.abs(duration_a - duration_b), max_value) / float(max_value)


def mean_area_distance(item_a, item_b, max_value):
    """
    Absolute difference in the means of the areas of each track over time.

    Args:
        item_a: STObject from the first set in TrackMatcher
        item_b: STObject from the second set in TrackMatcher
        max_value: Maximum distance value used as scaling value and upper constraint.

    Returns:
        Distance value between 0 and 1.
    """
    mean_area_a = np.mean([item_a.size(t) for t in item_a.times])
    mean_area_b = np.mean([item_b.size(t) for t in item_b.times])
    return np.abs(mean_area_a - mean_area_b) / float(max_value)



from pyproj import Proj
import numpy as np


def main():
    map_filename = '../ssef2015.map'
    proj_dict, grid_dict = read_arps_map_file(map_filename)
    print(proj_dict)
    print(grid_dict)
    print("Proj")
    mapping_data = make_proj_grids(proj_dict, grid_dict)
    print(mapping_data['lon'].shape)
    print(mapping_data['x'].shape)


def read_arps_map_file(map_filename):
    with open(map_filename) as map_file:
        map_params = map_file.read().split()
    proj_dict = {'proj': 'lcc', 'a': 6370000.0, 'b': 6370000.0, 'units': 'm'}
    grid_dict = {}
    proj_names = ['lat_2', 'lat_1', 'lat_0', 'lon_0']
    grid_names = ['sw_lat', 'sw_lon', 'ne_lat', 'ne_lon', 'dx', 'dy']
    for i in range(len(proj_names) + len(grid_names)):
        if i < len(proj_names):
            proj_dict[proj_names[i]] = float(map_params[i+2])
        else:
            j = i - len(proj_names)
            grid_dict[grid_names[j]] = float(map_params[i+2])
    return proj_dict, grid_dict


def read_ncar_map_file(map_filename):
    proj_keys = ["proj", "a", "b", "lat_2", "lat_1", "lat_0", "lon_0", "units"]
    grid_keys = ["sw_lat", "sw_lon", "ne_lat", "ne_lon", "dx", "dy"]
    proj_dict = {}
    grid_dict = {}
    with open(map_filename) as map_file:
        for line in map_file:
            map_option = line.split("=")
            if map_option[0] in ["a", "b", "lat_2", "lat_1", "lat_0", "lon_0",
                                 "sw_lat", "sw_lon", "ne_lat", "ne_lon", "dx", "dy"]:
                map_option[1] = float(map_option[1].strip())
            else:
                map_option[1] = map_option[1].strip()
            if map_option[0] in proj_keys:
                proj_dict[map_option[0]] = map_option[1]
            elif map_option[0] in grid_keys:
                grid_dict[map_option[0]] = map_option[1]
    return proj_dict, grid_dict


def make_proj_grids(proj_dict, grid_dict):
    map_proj = Proj(proj_dict)
    sw_x, sw_y = map_proj(grid_dict['sw_lon'], grid_dict['sw_lat'])
    ne_x, ne_y = map_proj(grid_dict['ne_lon'], grid_dict['ne_lat'])
    dx = grid_dict['dx']
    dy = grid_dict['dy']
    if proj_dict['units'] == "m":
        rounding = -2
    else:
        rounding = 0
    x = np.arange(np.round(sw_x, rounding), np.round(ne_x, rounding) + dx, dx)
    y = np.arange(np.round(sw_y, rounding), np.round(ne_y, rounding) + dy, dy)
    x_grid, y_grid = np.meshgrid(x, y)
    lon_grid, lat_grid = map_proj(x_grid, y_grid, inverse=True)
    mapping_data = {'lon': lon_grid, 'lat': lat_grid, 'x': x_grid, 'y': y_grid}
    return mapping_data


def get_proj_obj(proj_dict):
    return Proj(proj_dict)

if __name__ == "__main__":
    main()



from setuptools import setup
import os


classifiers = ['Development Status :: 4 - Beta',
               'Intended Audience :: Science/Research',
               'License :: OSI Approved :: MIT License',
               'Programming Language :: Python :: 2',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6',
               ]

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    requires = []
else:
    requires = ["numpy>=1.10",
                "pandas>=0.15",
                "scipy",
                "matplotlib>=1.5",
                "netCDF4",
                "scikit-learn>=0.16",
                "scikit-image",
                "pyproj",
                "pygrib",
                "arrow>=0.8.0",
                ]

if __name__ == "__main__":
    pkg_description = "Hagelslag is a Python package for storm-based analysis, forecasting, and evaluation."

    setup(name="hagelslag",
          version="0.3",
          description="Object-based severe weather forecast system",
          author="David John Gagne",
          author_email="djgagne@ou.edu",
          long_description=pkg_description,
          license="MIT",
          url="https://github.com/djgagne/hagelslag",
          packages=["hagelslag", "hagelslag.data", "hagelslag.processing", "hagelslag.evaluation", "hagelslag.util"],
          scripts=["bin/hsdata", "bin/hsforecast", "bin/hseval", "bin/hsplotter", "bin/hswrf3d", "bin/hsstation",
                   "bin/hsncarpatch","bin/hscalibration"],
          data_files=[("mapfiles", ["mapfiles/ssef2013.map", 
                                    "mapfiles/ssef2014.map", 
                                    "mapfiles/ssef2015.map", 
                                    "mapfiles/ncar_grib_table.txt",
                                    "mapfiles/hrrr_map_2016.txt",
                                    "mapfiles/ncar_ensemble_map_2015.txt",
                                    "mapfiles/ncar_2015_us_mask.nc",
                                    "mapfiles/ssef_2015_us_mask.nc"])],
          keywords=["hail", "verification", "tracking", "weather", "meteorology", "machine learning"],
          classifiers=classifiers,
          include_package_data=True,
          zip_safe=False,
          install_requires=requires)



import numpy as np
import pandas as pd
from hagelslag.data.MRMSGrid import MRMSGrid
from hagelslag.evaluation.ProbabilityMetrics import DistributedReliability, DistributedROC
from netCDF4 import Dataset
from datetime import timedelta
from scipy.signal import fftconvolve
from skimage.morphology import disk


class NeighborEvaluator(object):
    """
    A framework for statistically evaluating neighborhood probability forecasts.

    Attributes:
        run_date (datetime.datetime object): Date of the beginning of the model run
        start_hour (int): First forecast hour evaluated
        end_hour (int): Last forecast hour evaluated
        ensemble_name (str): Name of the ensemble system being evaluated
        model_name (str): Name of the physical or machine learning model being evaluated
        forecast_variable (str): Name of the forecast variable being evaluated.
        mrms_variable (str): Name of the NSSL MRMS product being used for gridded observations
        neighbor_radii (list or array): neighborhood radii in number of grid points
        smoothing_radii (list or array): radius of Gaussian filter used by the forecast
        obs_thresholds (list or array): Observed intensity threshold that corresponds with each element of
            size_thresholds
        size_thresholds (list or array): Intensity threshold for neighborhood probabilities
        obs_mask (bool): Whether or not another MRMS product is used to mask invalid grid points
        mask_variable (str): MRMS variable used for masking invalid grid points
        forecast_path (str): Path to forecast files
        mrms_path (str): Path to MRMS data
    """

    def __init__(self, run_date, start_hour, end_hour, ensemble_name, model_name, forecast_variable, mrms_variable,
                 neighbor_radii, smoothing_radii, obs_thresholds, size_thresholds, probability_levels, obs_mask,
                 mask_variable, forecast_path, mrms_path, coordinate_file=None, lon_bounds=None, lat_bounds=None):
        self.run_date = run_date
        self.start_hour = start_hour
        self.end_hour = end_hour
        self.ensemble_name = ensemble_name
        self.model_name = model_name
        self.forecast_variable = forecast_variable
        self.mrms_variable = mrms_variable
        self.obs_mask = obs_mask
        self.mask_variable = mask_variable
        self.neighbor_radii = neighbor_radii
        self.smoothing_radii = smoothing_radii
        self.obs_thresholds = obs_thresholds
        self.size_thresholds = size_thresholds
        self.probability_levels = probability_levels
        self.forecast_path = forecast_path
        self.mrms_path = mrms_path
        self.hourly_forecasts = {}
        self.period_forecasts = {}
        self.raw_obs = {}
        self.period_obs = {}
        self.coordinate_file = coordinate_file
        self.coordinates = {}
        self.lon_bounds = lon_bounds
        self.lat_bounds = lat_bounds

    def load_forecasts(self):
        """
        Load neighborhood probability forecasts.
        """
        run_date_str = self.run_date.strftime("%Y%m%d")
        forecast_file = self.forecast_path + "{0}/{1}_{2}_{3}_consensus_{0}.nc".format(run_date_str,
                                                                                       self.ensemble_name,
                                                                                       self.model_name,
                                                                                       self.forecast_variable)
        print("Forecast file: " + forecast_file)
        forecast_data = Dataset(forecast_file)
        for size_threshold in self.size_thresholds:
            for smoothing_radius in self.smoothing_radii:
                for neighbor_radius in self.neighbor_radii:
                    hour_var = "neighbor_prob_r_{0:d}_s_{1:d}_{2}_{3:0.2f}".format(neighbor_radius, smoothing_radius,
                                                                                   self.forecast_variable,
                                                                                   float(size_threshold))
                    period_var = "neighbor_prob_{0:d}-hour_r_{1:d}_s_{2:d}_{3}_{4:0.2f}".format(self.end_hour -
                                                                                                self.start_hour + 1,
                                                                                                neighbor_radius,
                                                                                                smoothing_radius,
                                                                                                self.forecast_variable,
                                                                                                float(size_threshold))

                    print("Loading forecasts {0} {1} {2} {3} {4}".format(self.run_date, self.model_name,
                                                                         self.forecast_variable, size_threshold,
                                                                         smoothing_radius))
                    if hour_var in forecast_data.variables.keys():
                        self.hourly_forecasts[hour_var] = forecast_data.variables[hour_var][:]
                    if period_var in forecast_data.variables.keys():
                        self.period_forecasts[period_var] = forecast_data.variables[period_var][:]
        forecast_data.close()

    def load_obs(self,  mask_threshold=0.5):
        """
        Loads observations and masking grid (if needed).

        Args:
            mask_threshold: Values greater than the threshold are kept, others are masked.
        """
        print("Loading obs ", self.run_date, self.model_name, self.forecast_variable)
        start_date = self.run_date + timedelta(hours=self.start_hour)
        end_date = self.run_date + timedelta(hours=self.end_hour)
        mrms_grid = MRMSGrid(start_date, end_date, self.mrms_variable, self.mrms_path)
        mrms_grid.load_data()
        if len(mrms_grid.data) > 0:
            self.raw_obs[self.mrms_variable] = np.where(mrms_grid.data > 100, 100, mrms_grid.data)
            self.period_obs[self.mrms_variable] = self.raw_obs[self.mrms_variable].max(axis=0)
            if self.obs_mask:
                mask_grid = MRMSGrid(start_date, end_date, self.mask_variable, self.mrms_path)
                mask_grid.load_data()
                self.raw_obs[self.mask_variable] = np.where(mask_grid.data >= mask_threshold, 1, 0)
                self.period_obs[self.mask_variable] = self.raw_obs[self.mask_variable].max(axis=0)

    def load_coordinates(self):
        """
        Loads lat-lon coordinates from a netCDF file.
        """
        coord_file = Dataset(self.coordinate_file)
        if "lon" in coord_file.variables.keys():
            self.coordinates["lon"] = coord_file.variables["lon"][:]
            self.coordinates["lat"] = coord_file.variables["lat"][:]
        else:
            self.coordinates["lon"] = coord_file.variables["XLONG"][0]
            self.coordinates["lat"] = coord_file.variables["XLAT"][0]
        coord_file.close()

    def evaluate_hourly_forecasts(self):
        """
        Calculates ROC curves and Reliability scores for each forecast hour.

        Returns:
            A pandas DataFrame containing forecast metadata as well as DistributedROC and Reliability objects.
        """
        score_columns = ["Run_Date", "Forecast_Hour", "Ensemble Name", "Model_Name", "Forecast_Variable",
                         "Neighbor_Radius", "Smoothing_Radius", "Size_Threshold", "ROC", "Reliability"]
        all_scores = pd.DataFrame(columns=score_columns)
        for h, hour in enumerate(range(self.start_hour, self.end_hour + 1)):
            for neighbor_radius in self.neighbor_radii:
                n_filter = disk(neighbor_radius)
                for s, size_threshold in enumerate(self.size_thresholds):
                    print("Eval hourly forecast {0:02d} {1} {2} {3} {4:d} {5:d}".format(hour, self.model_name,
                                                                                        self.forecast_variable,
                                                                                        self.run_date, neighbor_radius,
                                                                                        size_threshold))
                    hour_obs = fftconvolve(self.raw_obs[self.mrms_variable][h] >= self.obs_thresholds[s],
                                           n_filter, mode="same")
                    hour_obs[hour_obs > 1] = 1
                    hour_obs[hour_obs < 1] = 0
                    if self.obs_mask:
                        hour_obs = hour_obs[self.raw_obs[self.mask_variable][h] > 0]
                    for smoothing_radius in self.smoothing_radii:
                        hour_var = "neighbor_prob_r_{0:d}_s_{1:d}_{2}_{3:0.2f}".format(neighbor_radius,
                                                                                       smoothing_radius,
                                                                                       self.forecast_variable,
                                                                                       size_threshold)
                        if self.obs_mask:
                            hour_forecast = self.hourly_forecasts[hour_var][h][self.raw_obs[self.mask_variable][h] > 0]
                        else:
                            hour_forecast = self.hourly_forecasts[hour_var][h]
                        roc = DistributedROC(thresholds=self.probability_levels, obs_threshold=0.5)
                        roc.update(hour_forecast, hour_obs)
                        rel = DistributedReliability(thresholds=self.probability_levels, obs_threshold=0.5)
                        rel.update(hour_forecast, hour_obs)
                        row = [self.run_date, hour, self.ensemble_name, self.model_name, self.forecast_variable,
                               neighbor_radius,
                               smoothing_radius, size_threshold, roc, rel]
                        all_scores.loc[hour_var + "_{0:d}".format(hour)] = row
        return all_scores

    def evaluate_period_forecasts(self):
        """
        Evaluates ROC and Reliability scores for forecasts over the full period from start hour to end hour

        Returns:
            A pandas DataFrame with full-period metadata and verification statistics
        """
        score_columns = ["Run_Date", "Ensemble Name", "Model_Name", "Forecast_Variable", "Neighbor_Radius",
                         "Smoothing_Radius", "Size_Threshold",  "ROC", "Reliability"]
        all_scores = pd.DataFrame(columns=score_columns)
        if self.coordinate_file is not None:
            coord_mask = np.where((self.coordinates["lon"] >= self.lon_bounds[0]) &
                                  (self.coordinates["lon"] <= self.lon_bounds[1]) &
                                  (self.coordinates["lat"] >= self.lat_bounds[0]) &
                                  (self.coordinates["lat"] <= self.lat_bounds[1]) &
                                  (self.period_obs[self.mask_variable] > 0))
        else:
            coord_mask = None
        for neighbor_radius in self.neighbor_radii:
            n_filter = disk(neighbor_radius)
            for s, size_threshold in enumerate(self.size_thresholds):
                period_obs = fftconvolve(self.period_obs[self.mrms_variable] >= self.obs_thresholds[s],
                                         n_filter, mode="same")
                period_obs[period_obs > 1] = 1
                if self.obs_mask and self.coordinate_file is None:
                    period_obs = period_obs[self.period_obs[self.mask_variable] > 0]
                elif self.obs_mask and self.coordinate_file is not None:
                    period_obs = period_obs[coord_mask[0], coord_mask[1]]
                else:
                    period_obs = period_obs.ravel()
                for smoothing_radius in self.smoothing_radii:
                    print("Eval period forecast {0} {1} {2} {3} {4} {5}".format(self.model_name,
                                                                                self.forecast_variable,
                                                                                self.run_date,
                                                                                neighbor_radius,
                                                                                size_threshold, smoothing_radius))
                    period_var = "neighbor_prob_{0:d}-hour_r_{1:d}_s_{2:d}_{3}_{4:0.2f}".format(self.end_hour -
                                                                                                self.start_hour + 1,
                                                                                                neighbor_radius,
                                                                                                smoothing_radius,
                                                                                                self.forecast_variable,
                                                                                                size_threshold)
                    if self.obs_mask and self.coordinate_file is None:
                        period_forecast = self.period_forecasts[period_var][self.period_obs[self.mask_variable] > 0]
                    elif self.obs_mask and self.coordinate_file is not None:
                        period_forecast = self.period_forecasts[period_var][coord_mask[0], coord_mask[1]]
                    else:
                        period_forecast = self.period_forecasts[period_var].ravel()
                    roc = DistributedROC(thresholds=self.probability_levels, obs_threshold=0.5)
                    roc.update(period_forecast, period_obs)
                    rel = DistributedReliability(thresholds=self.probability_levels, obs_threshold=0.5)
                    rel.update(period_forecast, period_obs)
                    row = [self.run_date, self.ensemble_name, self.model_name, self.forecast_variable, neighbor_radius,
                           smoothing_radius, size_threshold, roc, rel]
                    all_scores.loc[period_var] = row
        return all_scores



import json
from multiprocessing import Pool
from scipy.stats import kstest, gamma
import numpy as np
import pandas as pd
from glob import glob
import matplotlib.pyplot as plt
import os 

def main():
    json_path = "/sharp/djgagne/track_data_spring2015_unique_json/"
    member = "wrf-s3m8_arw"
    run_dates = sorted(os.listdir(json_path))
    ks_frames = []
    pool = Pool(1)
    for run_date in run_dates:
        print(run_date)
        pool.apply_async(run_kstests, (json_path, run_date, member), callback=ks_frames.append)
    pool.close()
    pool.join()
    ks_total_frame = pd.concat(ks_frames)
    ks_total_frame.to_csv("ks_results_2015.csv", index_label="Id")
    return

def run_kstests(json_path, run_date, member):
    try:
        full_path = json_path + "/{0}/{1}/mesh_*.json".format(run_date, member)
        json_files = sorted(glob(full_path))
        ks_results = {"id":[], "ks":[]}
        for json_file in json_files:
            js = open(json_file)
            mesh_track = json.load(js)
            js.close()
            id = mesh_track["properties"]["id"]
            for m, mesh_obj in enumerate(mesh_track["features"]):
                step_id = id + "_{0:03d}".format(m)
                ts = np.array(mesh_obj["properties"]["timesteps"])
                mask = np.array(mesh_obj["properties"]["masks"])
                vals = ts[mask == 1]
                gdist = gamma.fit(vals, floc=vals.min()-0.1)
                sig = kstest(vals, gamma(*gdist).cdf)
                ks_results["id"].append(step_id)
                ks_results["ks"].append(sig)
                if sig[1] < 0.01:
                    print(step_id,)
                    print(sig[1],gdist)
                    print(np.sort(vals))
                    plt.figure(figsize=(8,8))
                    plt.pcolormesh(ts, alpha=0.5, cmap="YlOrRd", vmin=0, vmax=100)
                    pc = plt.pcolormesh(np.ma.array(ts, mask=mask==0), cmap="YlOrRd", vmin=0, vmax=100)
                    plt.title(step_id)
                    plt.colorbar(pc)
                    plt.savefig(step_id + ".png", bbox_inches="tight", dpi=150)
                    plt.close()
        ks_frame = pd.DataFrame(ks_results["ks"], index=ks_results["id"],columns=["D", "p-val"])
        print(ks_frame.shape[0])
    except Exception as e:
        raise e
    return ks_frame
if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
from hagelslag.data.ModelGrid import ModelGrid
from os.path import join


class HRRRModelGrid(ModelGrid):
    def __init__(self, run_date, variable, start_date, end_date, path, frequency="1H"):
        self.run_date = pd.Timestamp(run_date)
        self.variable = variable
        self.start_date = pd.Timestamp(start_date)
        self.end_date = pd.Timestamp(end_date)
        self.frequency = frequency
        self.path = path
        self.valid_dates = pd.DatetimeIndex(start=self.start_date,
                                            end=self.end_date,
                                            freq=self.frequency)
        self.forecast_hours = np.array((self.valid_dates - self.run_date).total_seconds() / 3600, dtype=int)
        filenames = []
        for forecast_hour in self.forecast_hours:
            filenames.append(join(self.path, self.run_date.strftime("%Y%m%d"),
                                  "{0}_f{1:03d}_HRRR.nc4".format(self.run_date.strftime("%Y%m%d_i%H"),
                                                                 forecast_hour)))

        super(HRRRModelGrid, self).__init__(filenames, run_date, start_date, end_date, variable)





import numpy as np
import pandas as pd
import pickle
import json
import os
from sklearn.decomposition import PCA
from copy import deepcopy
from glob import glob
from multiprocessing import Pool
from sklearn.linear_model import LinearRegression
from hagelslag.evaluation.ProbabilityMetrics import DistributedROC
from os.path import join
from hagelslag.util.make_proj_grids import read_arps_map_file, read_ncar_map_file

try:
    from sklearn.model_selection import KFold
except ImportError:
    from sklearn.cross_validation import KFold


class TrackModeler(object):
    """
    TrackModeler is designed to load and process data generated by TrackProcessing and then use that data to fit
    machine learning models to predict whether or not hail will occur, hail size, and translation errors in time and
    space.
    """

    def __init__(self,
                 ensemble_name,
                 train_data_path,
                 forecast_data_path,
                 member_files,
                 start_dates,
                 end_dates,
                 sector,
                 map_file,
                 group_col="Microphysics"):
        self.train_data_path = train_data_path
        self.forecast_data_path = forecast_data_path
        self.ensemble_name = ensemble_name
        self.member_files = member_files
        self.start_dates = start_dates
        self.end_dates = end_dates
        self.data = {"train": {}, "forecast": {}}
        self.condition_models = {}
        self.size_models = {}
        self.size_distribution_models = {}
        self.track_models = {"translation-x": {},
                             "translation-y": {},
                             "start-time": {}}
        self.sector = sector
        self.map_file = map_file
        self.group_col = group_col
            
        if self.map_file[-3:] == "map":                                  
            self.proj_dict, self.grid_dict = read_arps_map_file(self.map_file)
        else:                                               
            self.proj_dict, self.grid_dict = read_ncar_map_file(self.map_file)   
            
        
        return

    def load_data(self, mode="train", format="csv"):
        """
        Load data from flat data files containing total track information and information about each timestep.
        The two sets are combined using merge operations on the Track IDs. Additional member information is gathered
        from the appropriate member file.
        Args:
            mode: "train" or "forecast"
            format:  file format being used. Default is "csv"
        """
        if mode in self.data.keys():
            run_dates = pd.DatetimeIndex(start=self.start_dates[mode],
                                        end=self.end_dates[mode],freq="1D")
            run_date_str = [d.strftime("%Y%m%d-%H%M") for d in run_dates.date]
            print(run_date_str)
            all_total_track_files = sorted(glob(getattr(self, mode + "_data_path") +
                                                "*total_" + self.ensemble_name + "*." + format))
            all_step_track_files = sorted(glob(getattr(self, mode + "_data_path") +
                                               "*step_" + self.ensemble_name + "*." + format))
            total_track_files = []
            for track_file in all_total_track_files:
                file_date = track_file.split("_")[-1][:-4]
                if file_date in run_date_str:
                    total_track_files.append(track_file)
            step_track_files = []
            for step_file in all_step_track_files:
                file_date = step_file.split("_")[-1][:-4]
                if file_date in run_date_str:
                    step_track_files.append(step_file)            
	  	   
            self.data[mode]["total"] = pd.concat(map(pd.read_csv, total_track_files),
                                                 ignore_index=True)
            self.data[mode]["total"] = self.data[mode]["total"].fillna(value=0)
            self.data[mode]["total"] = self.data[mode]["total"].replace([np.inf, -np.inf], 0)
            self.data[mode]["step"] = pd.concat(map(pd.read_csv, step_track_files),
                                                ignore_index=True)
            self.data[mode]["step"] = self.data[mode]["step"].fillna(value=0)
            self.data[mode]["step"] = self.data[mode]["step"].replace([np.inf, -np.inf], 0)
            if mode == "forecast":
                self.data[mode]["step"] = self.data[mode]["step"].drop_duplicates("Step_ID")
            self.data[mode]["member"] = pd.read_csv(self.member_files[mode])
            self.data[mode]["combo"] = pd.merge(self.data[mode]["step"],
                                                self.data[mode]["total"],
                                                on=["Track_ID", "Ensemble_Name", "Ensemble_Member", "Run_Date"])
            self.data[mode]["combo"] = pd.merge(self.data[mode]["combo"],
                                                self.data[mode]["member"],
                                                on="Ensemble_Member") 
            self.data[mode]["total_group"] = pd.merge(self.data[mode]["total"],
                                                      self.data[mode]["member"],
                                                      on="Ensemble_Member")

    def calc_copulas(self,
                     output_file,
                     model_names=("start-time", "translation-x", "translation-y"),
                     label_columns=("Start_Time_Error", "Translation_Error_X", "Translation_Error_Y")):
        """
        Calculate a copula multivariate normal distribution from the training data for each group of ensemble members.
        Distributions are written to a pickle file for later use.
        Args:
            output_file: Pickle file
            model_names: Names of the tracking models
            label_columns: Names of the data columns used for labeling
        Returns:
        """
        if len(self.data['train']) == 0:
            self.load_data()
        groups = self.data["train"]["member"][self.group_col].unique()
        copulas = {}
        label_columns = list(label_columns)
        for group in groups:
            print(group)
            group_data = self.data["train"]["total_group"].loc[
                self.data["train"]["total_group"][self.group_col] == group]
            group_data = group_data.dropna()
            group_data.reset_index(drop=True, inplace=True)
            copulas[group] = {}
            copulas[group]["mean"] = group_data[label_columns].mean(axis=0).values
            copulas[group]["cov"] = np.cov(group_data[label_columns].values.T)
            copulas[group]["model_names"] = list(model_names)
            del group_data
        pickle.dump(copulas, open(output_file, "w"), pickle.HIGHEST_PROTOCOL)

    def fit_condition_models(self, model_names,
                             model_objs,
                             input_columns,
                             output_column="Matched",
                             output_threshold=0.0):
        """
        Fit machine learning models to predict whether or not hail will occur.
        Args:
            model_names: List of strings with the names for the particular machine learning models
            model_objs: scikit-learn style machine learning model objects.
            input_columns: list of the names of the columns used as input for the machine learning model
            output_column: name of the column used for labeling whether or not the event occurs
            output_threshold: splitting threshold to determine if event has occurred. Default 0.0
        """
        print("Fitting condition models")
        groups = self.data["train"]["member"][self.group_col].unique()
        
        weights = None

        for group in groups:
            print(group)
            group_data = self.data["train"]["combo"].loc[self.data["train"]["combo"][self.group_col] == group] 
            if self.sector:
        
                lon_obj = data.loc[:,'Centroid_Lon']
                lat_obj = data.loc[:,'Centroid_Lat']

                left_lon,right_lon = self.grid_dict["sw_lon"],self.grid_dict["ne_lon"]
                lower_lat,upper_lat = self.grid_dict["sw_lat"],self.grid_dict["ne_lat"]

                weights = np.where((left_lon<=lon_obj)&(right_lon>=lon_obj) &\
                    (lower_lat<=lat_obj)&(upper_lat>=lat_obj),1,0.3)
            
            output_data = np.where(group_data[output_column] > output_threshold, 1, 0)
            print("Ones: ", np.count_nonzero(output_data > 0), "Zeros: ", np.count_nonzero(output_data == 0))
            self.condition_models[group] = {}
            for m, model_name in enumerate(model_names):
                print(model_name)
                self.condition_models[group][model_name] = deepcopy(model_objs[m])
                try:
                    self.condition_models[group][model_name].fit(group_data[input_columns], 
                                output_data,sample_weight=weights)
                except:
                    self.condition_models[group][model_name].fit(group_data[input_columns], output_data)

                if hasattr(self.condition_models[group][model_name], "best_estimator_"):
                    print(self.condition_models[group][model_name].best_estimator_)
                    print(self.condition_models[group][model_name].best_score_)

    def fit_condition_threshold_models(self, model_names, model_objs, input_columns, output_column="Matched",
                                       output_threshold=0.5, num_folds=5, threshold_score="ets"):
        """
        Fit models to predict hail/no-hail and use cross-validation to determine the probaility threshold that
        maximizes a skill score.

        Args:
            model_names: List of machine learning model names
            model_objs: List of Scikit-learn ML models
            input_columns: List of input variables in the training data
            output_column: Column used for prediction
            output_threshold: Values exceeding this threshold are considered positive events; below are nulls
            num_folds: Number of folds in the cross-validation procedure
            threshold_score: Score available in ContingencyTable used for determining the best probability threshold
        Returns:
            None
        """
        print("Fitting condition models")
        groups = self.data["train"]["member"][self.group_col].unique()
        
        weights=None

        for group in groups:
            print(group)
            group_data = self.data["train"]["combo"].iloc[
                np.where(self.data["train"]["combo"][self.group_col] == group)[0]]
            
            if self.sector:
                lon_obj = group_data.loc[:,'Centroid_Lon']
                lat_obj = group_data.loc[:,'Centroid_Lat']
                
                conus_lat_lon_points = zip(lon_obj.values.ravel(),lat_obj.values.ravel())
                center_lon, center_lat = self.proj_dict["lon_0"],self.proj_dict["lat_0"] 
            
                distances = np.array([np.sqrt((x-center_lon)**2+\
                        (y-center_lat)**2) for (x, y) in conus_lat_lon_points])
            
                min_dist, max_minus_min = min(distances),max(distances)-min(distances)

                distance_0_1 = [1.0-((d - min_dist)/(max_minus_min)) for d in distances]
                weights = np.array(distance_0_1)
        
            output_data = np.where(group_data.loc[:, output_column] > output_threshold, 1, 0)
            ones = np.count_nonzero(output_data > 0)
            print("Ones: ", ones, "Zeros: ", np.count_nonzero(output_data == 0))
            self.condition_models[group] = {}
            num_elements = group_data[input_columns].shape[0]
            
            for m, model_name in enumerate(model_names):
                print(model_name)    
                roc = DistributedROC(thresholds=np.arange(0, 1.1, 0.01))
                self.condition_models[group][model_name] = deepcopy(model_objs[m])

                try:
                    kf = KFold(n_splits=num_folds)
                    for train_index, test_index in kf.split(group_data[input_columns].values):
                        if np.count_nonzero(output_data[train_index]) > 0:
                            try:
                                self.condition_models[group][model_name].fit(
                                        group_data.iloc[train_index][input_columns],
                                        output_data[train_index],sample_weight=weights[train_index])
                            except:
                                self.condition_models[group][model_name].fit(
                                        group_data.iloc[train_index][input_columns],
                                        output_data[train_index])
                            
                            cv_preds = self.condition_models[group][model_name].predict_proba(
                                group_data.iloc[test_index][input_columns])[:,1]
                            
                            roc.update(cv_preds, output_data[test_index])
                        
                        else:
                            continue

                except TypeError:
                    kf = KFold(num_elements,n_folds=num_folds)
                    for train_index, test_index in kf:

                        if np.count_nonzero(output_data[train_index]) > 0:
                            try:
                                self.condition_models[group][model_name].fit(
                                        group_data.iloc[train_index][input_columns],
                                        output_data[train_index],sample_weight=weights[train_index])
                            except:
                                self.condition_models[group][model_name].fit(
                                        group_data.iloc[train_index][input_columns],
                                        output_data[train_index])
                            cv_preds = self.condition_models[group][model_name].predict_proba(
                                group_data.iloc[test_index][input_columns])[:, 1]
                            
                            roc.update(cv_preds, output_data[test_index])
                        
                        else:
                            continue

                self.condition_models[group][
                    model_name + "_condition_threshold"], _ = roc.max_threshold_score(threshold_score)
                print(model_name + " condition threshold: {0:0.3f}".format(
                self.condition_models[group][model_name + "_condition_threshold"]))
                self.condition_models[group][model_name].fit(group_data[input_columns], output_data)

    def predict_condition_models(self, model_names,
                                 input_columns,
                                 metadata_cols,
                                 data_mode="forecast",
                                 ):
        """
        Apply condition modelsto forecast data.
        Args:
            model_names: List of names associated with each condition model used for prediction
            input_columns: List of columns in data used as input into the model
            metadata_cols: Columns from input data that should be included in the data frame with the predictions.
            data_mode: Which data subset to pull from for the predictions, "forecast" by default
        Returns:
            A dictionary of data frames containing probabilities of the event and specified metadata
        """
        groups = self.condition_models.keys()
        predictions = pd.DataFrame(self.data[data_mode]["combo"][metadata_cols])
        for group in groups:
            print(group)
            print(self.condition_models[group])
            g_idxs = self.data[data_mode]["combo"][self.group_col] == group
            group_count = np.count_nonzero(g_idxs)
            if group_count > 0:
                for m, model_name in enumerate(model_names):
                    mn = model_name.replace(" ", "-")
                    predictions.loc[g_idxs, mn + "_conditionprob"] = self.condition_models[group][
                                                                         model_name].predict_proba(
                        self.data[data_mode]["combo"].loc[g_idxs, input_columns])[:, 1]
                    predictions.loc[g_idxs,
                                    mn + "_conditionthresh"] = np.where(predictions.loc[g_idxs, mn + "_conditionprob"]
                                                                        >= self.condition_models[group][
                                                                            model_name + "_condition_threshold"], 1, 0)

        return predictions

    def fit_size_distribution_models(self, model_names, model_objs, input_columns,
                                     output_columns=None, calibrate=False):
        """
        Fits multitask machine learning models to predict the parameters of a size distribution
        Args:
            model_names: List of machine learning model names
            model_objs: scikit-learn style machine learning model objects
            input_columns: Training data columns used as input for ML model
            output_columns: Training data columns used for prediction
            calibrate: Whether or not to fit a log-linear regression to predictions from ML model
        """
        if output_columns is None:
            output_columns = ["Shape", "Location", "Scale"]
        groups = np.unique(self.data["train"]["member"][self.group_col])
        
        weights=None
        
        for group in groups:
            group_data = self.data["train"]["combo"].loc[self.data["train"]["combo"][self.group_col] == group]
            group_data = group_data.dropna()
            group_data = group_data[group_data[output_columns[-1]] > 0]
            if self.sector:
        
                lon_obj = group_data.loc[:,'Centroid_Lon']
                lat_obj = group_data.loc[:,'Centroid_Lat']
                
                conus_lat_lon_points = zip(lon_obj.values.ravel(),lat_obj.values.ravel())
                center_lon, center_lat = self.proj_dict["lon_0"],self.proj_dict["lat_0"] 
            
                distances = np.array([np.sqrt((x-center_lon)**2+\
                        (y-center_lat)**2) for (x, y) in conus_lat_lon_points])
            
                min_dist, max_minus_min = min(distances),max(distances)-min(distances)

                distance_0_1 = [1.0-((d - min_dist)/(max_minus_min)) for d in distances]
                weights = np.array(distance_0_1)

            self.size_distribution_models[group] = {"multi": {}, "lognorm": {}}
            if calibrate:
                self.size_distribution_models[group]["calshape"] = {}
                self.size_distribution_models[group]["calscale"] = {}
            log_labels = np.log(group_data[output_columns].values)
            log_means = log_labels.mean(axis=0)
            log_sds = log_labels.std(axis=0)
            self.size_distribution_models[group]['lognorm']['mean'] = log_means
            self.size_distribution_models[group]['lognorm']['sd'] = log_sds
            for m, model_name in enumerate(model_names):
                print(group, model_name)
                self.size_distribution_models[group]["multi"][model_name] = deepcopy(model_objs[m])
                try:
                    self.size_distribution_models[group]["multi"][model_name].fit(group_data[input_columns],
                                                                              (log_labels - log_means) / log_sds,
                                                                        sample_weight=weights)
                except:
                    self.size_distribution_models[group]["multi"][model_name].fit(group_data[input_columns],
                                                                              (log_labels - log_means) / log_sds)
                if calibrate:
                    training_predictions = self.size_distribution_models[
                        group]["multi"][model_name].predict(group_data[input_columns])
                    self.size_distribution_models[group]["calshape"][model_name] = LinearRegression()
                    self.size_distribution_models[group]["calshape"][model_name].fit(training_predictions[:, 0:1],
                                                                                     (log_labels[:, 0] - log_means[0]) /
                                                                                     log_sds[
                                                                                         0],
                                                                                    sample_weight=weights)
                    self.size_distribution_models[group]["calscale"][model_name] = LinearRegression()
                    self.size_distribution_models[group]["calscale"][model_name].fit(training_predictions[:, 1:],
                                                                                     (log_labels[:, 1] - log_means[1]) /
                                                                                     log_sds[
                                                                                         1],
                                                                            sample_weight=weights)


    def fit_size_distribution_component_models(self, model_names, model_objs, input_columns, output_columns):
        """
        This calculates 2 principal components for the hail size distribution between the shape and scale parameters.
        Separate machine learning models are fit to predict each component.

        Args:
            model_names: List of machine learning model names
            model_objs: List of machine learning model objects.
            input_columns: List of input variables
            output_columns: Output columns, should contain Shape and Scale.

        Returns:

        """
        groups = np.unique(self.data["train"]["member"][self.group_col])
        
        weights=None

        for group in groups:
            print(group)
            group_data = self.data["train"]["combo"].loc[self.data["train"]["combo"][self.group_col] == group]
            group_data = group_data.dropna()
            group_data = group_data.loc[group_data[output_columns[-1]] > 0]
            
            if self.sector:
        
                lon_obj = group_data.loc[:,'Centroid_Lon']
                lat_obj = group_data.loc[:,'Centroid_Lat']
                
                conus_lat_lon_points = zip(lon_obj.values.ravel(),lat_obj.values.ravel())
                center_lon, center_lat = self.proj_dict["lon_0"],self.proj_dict["lat_0"] 
            
                distances = np.array([np.sqrt((x-center_lon)**2+\
                        (y-center_lat)**2) for (x, y) in conus_lat_lon_points])
            
                min_dist, max_minus_min = min(distances),max(distances)-min(distances)

                distance_0_1 = [1.0-((d - min_dist)/(max_minus_min)) for d in distances]
                weights = np.array(distance_0_1)

            
            self.size_distribution_models[group] = {"lognorm": {}}
            self.size_distribution_models[group]["lognorm"]["pca"] = PCA(n_components=len(output_columns))
            log_labels = np.log(group_data[output_columns].values)
            log_labels[:, np.where(output_columns == "Shape")[0]] *= -1
            log_means = log_labels.mean(axis=0)
            log_sds = log_labels.std(axis=0)
            log_norm_labels = (log_labels - log_means) / log_sds
            out_pc_labels = self.size_distribution_models[group]["lognorm"]["pca"].fit_transform(log_norm_labels)
            self.size_distribution_models[group]['lognorm']['mean'] = log_means
            self.size_distribution_models[group]['lognorm']['sd'] = log_sds
            for comp in range(len(output_columns)):
                self.size_distribution_models[group]["pc_{0:d}".format(comp)] = dict()
                for m, model_name in enumerate(model_names):
                    print(model_name, comp)
                    self.size_distribution_models[group][
                        "pc_{0:d}".format(comp)][model_name] = deepcopy(model_objs[m])
                    try:
                        self.size_distribution_models[group][
                            "pc_{0:d}".format(comp)][model_name].fit(group_data[input_columns],
                                                                 out_pc_labels[:, comp],
                                                            sample_weight=weights)
                    except:
                        self.size_distribution_models[group][
                            "pc_{0:d}".format(comp)][model_name].fit(group_data[input_columns],
                                                                 out_pc_labels[:, comp])
        return

    def predict_size_distribution_models(self, model_names, input_columns, metadata_cols,
                                         data_mode="forecast", location=6, calibrate=False):
        """
        Make predictions using fitted size distribution models.
        Args:
            model_names: Name of the models for predictions
            input_columns: Data columns used for input into ML models
            metadata_cols: Columns from input data that should be included in the data frame with the predictions.
            data_mode: Set of data used as input for prediction models
            location: Value of fixed location parameter
            calibrate: Whether or not to apply calibration model
        Returns:
            Predictions in dictionary of data frames grouped by group type
        """
        groups = self.size_distribution_models.keys()
        predictions = {}
        for group in groups:
            group_data = self.data[data_mode]["combo"].loc[self.data[data_mode]["combo"][self.group_col] == group]
            predictions[group] = group_data[metadata_cols]
            if group_data.shape[0] > 0:
                log_mean = self.size_distribution_models[group]["lognorm"]["mean"]
                log_sd = self.size_distribution_models[group]["lognorm"]["sd"]
                for m, model_name in enumerate(model_names):
                    multi_predictions = self.size_distribution_models[group]["multi"][model_name].predict(
                        group_data[input_columns])
                    if calibrate:
                        multi_predictions[:, 0] = self.size_distribution_models[group]["calshape"][model_name].predict(
                            multi_predictions[:, 0:1])
                        multi_predictions[:, 1] = self.size_distribution_models[group]["calscale"][model_name].predict(
                            multi_predictions[:, 1:])
                    multi_predictions = np.exp(multi_predictions * log_sd + log_mean)
                    if multi_predictions.shape[1] == 2:
                        multi_predictions_temp = np.zeros((multi_predictions.shape[0], 3))
                        multi_predictions_temp[:, 0] = multi_predictions[:, 0]
                        multi_predictions_temp[:, 1] = location
                        multi_predictions_temp[:, 2] = multi_predictions[:, 1]
                        multi_predictions = multi_predictions_temp
                    for p, pred_col in enumerate(["shape", "location", "scale"]):
                        predictions[group][model_name].loc[:, model_name.replace(" ", "-") + "_" + pred_col] = \
                            multi_predictions[:, p]
        return predictions

    def predict_size_distribution_component_models(self, model_names, input_columns, output_columns, metadata_cols,
                                                   data_mode="forecast", location=6):
        """
        Make predictions using fitted size distribution models.
        Args:
            model_names: Name of the models for predictions
            input_columns: Data columns used for input into ML models
            output_columns: Names of output columns
            metadata_cols: Columns from input data that should be included in the data frame with the predictions.
            data_mode: Set of data used as input for prediction models
            location: Value of fixed location parameter
        Returns:
            Predictions in dictionary of data frames grouped by group type
        """
        groups = self.size_distribution_models.keys()
        predictions = pd.DataFrame(self.data[data_mode]["combo"][metadata_cols])
        for group in groups:
            group_idxs = self.data[data_mode]["combo"][self.group_col] == group
            group_count = np.count_nonzero(group_idxs)
            print(self.size_distribution_models[group])
            if group_count > 0:
                log_mean = self.size_distribution_models[group]["lognorm"]["mean"]
                log_sd = self.size_distribution_models[group]["lognorm"]["sd"]
                for m, model_name in enumerate(model_names):
                    raw_preds = np.zeros((group_count, len(output_columns)))
                    for c in range(len(output_columns)):
                        raw_preds[:, c] = self.size_distribution_models[group][
                            "pc_{0:d}".format(c)][model_name].predict(self.data[data_mode]["combo"].loc[group_idxs,
                                                                                                        input_columns])
                    log_norm_preds = self.size_distribution_models[group]["lognorm"]["pca"].inverse_transform(raw_preds)
                    log_norm_preds[:, 0] *= -1
                    multi_predictions = np.exp(log_norm_preds * log_sd + log_mean)
                    if multi_predictions.shape[1] == 2:
                        multi_predictions_temp = np.zeros((multi_predictions.shape[0], 3))
                        multi_predictions_temp[:, 0] = multi_predictions[:, 0]
                        multi_predictions_temp[:, 1] = location
                        multi_predictions_temp[:, 2] = multi_predictions[:, 1]
                        multi_predictions = multi_predictions_temp
                    for p, pred_col in enumerate(["shape", "location", "scale"]):
                        predictions.loc[group_idxs, model_name.replace(" ", "-") + "_" + pred_col] = \
                            multi_predictions[:, p]
        return predictions

    def fit_size_models(self, model_names,
                        model_objs,
                        input_columns,
                        output_column="Hail_Size",
                        output_start=5,
                        output_step=5,
                        output_stop=100):
        """
        Fit size models to produce discrete pdfs of forecast hail sizes.
        Args:
            model_names: List of model names
            model_objs: List of model objects
            input_columns: List of input variables
            output_column: Output variable name
            output_start: Hail size bin start
            output_step: hail size bin step
            output_stop: hail size bin stop
        """
        print("Fitting size models")
        groups = self.data["train"]["member"][self.group_col].unique()
        output_start = int(output_start)
        output_step = int(output_step)
        output_stop = int(output_stop)
        for group in groups:
            group_data = self.data["train"]["combo"].loc[self.data["train"]["combo"][self.group_col] == group]
            group_data.dropna(inplace=True)
            group_data = group_data[group_data[output_column] >= output_start]
            output_data = group_data[output_column].values.astype(int)
            output_data[output_data > output_stop] = output_stop
            discrete_data = ((output_data - output_start) // output_step) * output_step + output_start
            self.size_models[group] = {}
            self.size_models[group]["outputvalues"] = np.arange(output_start, output_stop + output_step, output_step,
                                                                dtype=int)
            for m, model_name in enumerate(model_names):
                print("{0} {1}".format(group, model_name))
                self.size_models[group][model_name] = deepcopy(model_objs[m])
                self.size_models[group][model_name].fit(group_data[input_columns], discrete_data)

    def predict_size_models(self, model_names,
                            input_columns,
                            metadata_cols,
                            data_mode="forecast"):
        """
        Apply size models to forecast data.
        Args:
            model_names:
            input_columns:
            metadata_cols:
            data_mode:
        """
        groups = self.size_models.keys()
        predictions = {}
        for group in groups:
            group_data = self.data[data_mode]["combo"].loc[self.data[data_mode]["combo"][self.group_col] == group]
            if group_data.shape[0] > 0:
                predictions[group] = {}
                output_values = self.size_models[group]["outputvalues"].astype(int)
                for m, model_name in enumerate(model_names):
                    print("{0} {1}".format(group, model_name))
                    pred_col_names = [model_name.replace(" ", "-") + "_{0:02d}".format(p) for p in output_values]
                    predictions[group][model_name] = group_data[metadata_cols]
                    pred_vals = self.size_models[group][model_name].predict_proba(group_data[input_columns])
                    pred_classes = self.size_models[group][model_name].classes_
                    pred_pdf = np.zeros((pred_vals.shape[0], output_values.size))
                    for pcv, pc in enumerate(pred_classes):
                        idx = np.where(output_values == pc)[0][0]
                        pred_pdf[:, idx] = pred_vals[:, pcv]
                    for pcn, pred_col_name in enumerate(pred_col_names):
                        predictions[group][model_name].loc[:, pred_col_name] = pred_pdf[:, pcn]
        return predictions

    def fit_track_models(self,
                         model_names,
                         model_objs,
                         input_columns,
                         output_columns,
                         output_ranges,
                         ):
        """
        Fit machine learning models to predict track error offsets.
            model_names:
            model_objs:
            input_columns:
            output_columns:
            output_ranges:
        """
        print("Fitting track models")
        groups = self.data["train"]["member"][self.group_col].unique()
        for group in groups:
            group_data = self.data["train"]["combo"].loc[self.data["train"]["combo"][self.group_col] == group]
            group_data = group_data.dropna()
            group_data = group_data.loc[group_data["Duration_Step"] == 1]
            for model_type, model_dict in self.track_models.items():
                model_dict[group] = {}
                output_data = group_data[output_columns[model_type]].values.astype(int)
                output_data[output_data < output_ranges[model_type][0]] = output_ranges[model_type][0]
                output_data[output_data > output_ranges[model_type][1]] = output_ranges[model_type][1]
                discrete_data = (output_data - output_ranges[model_type][0]) // output_ranges[model_type][2] * \
                                output_ranges[model_type][2] + output_ranges[model_type][0]
                model_dict[group]["outputvalues"] = np.arange(output_ranges[model_type][0],
                                                              output_ranges[model_type][1] +
                                                              output_ranges[model_type][2],
                                                              output_ranges[model_type][2])
                for m, model_name in enumerate(model_names):
                    print("{0} {1} {2}".format(group, model_type, model_name))
                    model_dict[group][model_name] = deepcopy(model_objs[m])
                    model_dict[group][model_name].fit(group_data[input_columns], discrete_data)

    def predict_track_models(self, model_names,
                             input_columns,
                             metadata_cols,
                             data_mode="forecast",
                             ):
        """
        Predict track offsets on forecast data.
        Args:
            model_names:
            input_columns:
            metadata_cols:
            data_mode:
        Returns:
        """
        predictions = {}
        for model_type, track_model_set in self.track_models.items():
            predictions[model_type] = {}
            groups = track_model_set.keys()
            for group in groups:
                group_data = self.data[data_mode]["combo"].loc[self.data[data_mode]["combo"][self.group_col] == group]
                if group_data.shape[0] > 0:
                    predictions[model_type][group] = {}
                    output_values = track_model_set[group]["outputvalues"].astype(int)
                    for m, model_name in enumerate(model_names):
                        print("{0} {1} {2}".format(group, model_type, model_name))
                        pred_col_names = [model_name.replace(" ", "-") + "_{0:02d}".format(p) for p in output_values]
                        predictions[model_type][group][model_name] = group_data[metadata_cols]
                        pred_vals = track_model_set[group][model_name].predict_proba(group_data[input_columns])
                        pred_classes = track_model_set[group][model_name].classes_
                        pred_pdf = np.zeros((pred_vals.shape[0], output_values.size))
                        for pcv, pc in enumerate(pred_classes):
                            idx = np.where(pc == output_values)[0][0]
                            pred_pdf[:, idx] = pred_vals[:, pcv]
                        for pcn, pred_col_name in enumerate(pred_col_names):
                            predictions[model_type][group][model_name].loc[:, pred_col_name] = pred_pdf[:, pcn]
        return predictions

    def save_models(self, model_path):
        """
        Save machine learning models to pickle files.
        """
        for group, condition_model_set in self.condition_models.items():
            for model_name, model_obj in condition_model_set.items():
                out_filename = model_path + \
                               "{0}_{1}_condition.pkl".format(group,
                                                              model_name.replace(" ", "-"))
                with open(out_filename, "wb") as pickle_file:
                    pickle.dump(model_obj,
                                pickle_file,
                                pickle.HIGHEST_PROTOCOL)
        for group, size_model_set in self.size_models.items():
            for model_name, model_obj in size_model_set.items():
                out_filename = model_path + \
                               "{0}_{1}_size.pkl".format(group,
                                                         model_name.replace(" ", "-"))
                with open(out_filename, "wb") as pickle_file:
                    pickle.dump(model_obj,
                                pickle_file,
                                pickle.HIGHEST_PROTOCOL)
        for group, dist_model_set in self.size_distribution_models.items():
            for model_type, model_objs in dist_model_set.items():
                for model_name, model_obj in model_objs.items():
                    out_filename = model_path + \
                                   "{0}_{1}_{2}_sizedist.pkl".format(group,
                                                                     model_name.replace(" ", "-"),
                                                                     model_type)
                    with open(out_filename, "wb") as pickle_file:
                        pickle.dump(model_obj,
                                    pickle_file,
                                    pickle.HIGHEST_PROTOCOL)
        for model_type, track_type_models in self.track_models.items():
            for group, track_model_set in track_type_models.items():
                for model_name, model_obj in track_model_set.items():
                    out_filename = model_path + \
                                   "{0}_{1}_{2}_track.pkl".format(group,
                                                                  model_name.replace(" ", "-"),
                                                                  model_type)
                    with open(out_filename, "wb") as pickle_file:
                        pickle.dump(model_obj,
                                    pickle_file,
                                    pickle.HIGHEST_PROTOCOL)

        return

    def load_models(self, model_path):
        """
        Load models from pickle files.
        """
        condition_model_files = sorted(glob(model_path + "*_condition.pkl"))
        if len(condition_model_files) > 0:
            for condition_model_file in condition_model_files:
                model_comps = condition_model_file.split("/")[-1][:-4].split("_")
                if model_comps[0] not in self.condition_models.keys():
                    self.condition_models[model_comps[0]] = {}
                model_name = model_comps[1].replace("-", " ")
                with open(condition_model_file, "rb") as cmf:
                    if "condition_threshold" in condition_model_file:
                        self.condition_models[model_comps[0]][model_name + "_condition_threshold"] = pickle.load(cmf)
                    else:
                        self.condition_models[model_comps[0]][model_name] = pickle.load(cmf)

        size_model_files = sorted(glob(model_path + "*_size.pkl"))
        if len(size_model_files) > 0:
            for size_model_file in size_model_files:
                model_comps = size_model_file.split("/")[-1][:-4].split("_")
                if model_comps[0] not in self.size_models.keys():
                    self.size_models[model_comps[0]] = {}
                model_name = model_comps[1].replace("-", " ")
                with open(size_model_file, "rb") as smf:
                    self.size_models[model_comps[0]][model_name] = pickle.load(smf)

        size_dist_model_files = sorted(glob(model_path + "*_sizedist.pkl"))
        if len(size_dist_model_files) > 0:
            for dist_model_file in size_dist_model_files:
                model_comps = dist_model_file.split("/")[-1][:-4].split("_")
                if model_comps[0] not in self.size_distribution_models.keys():
                    self.size_distribution_models[model_comps[0]] = {}
                if "_".join(model_comps[2:-1]) not in self.size_distribution_models[model_comps[0]].keys():
                    self.size_distribution_models[model_comps[0]]["_".join(model_comps[2:-1])] = {}
                model_name = model_comps[1].replace("-", " ")
                with open(dist_model_file, "rb") as dmf:
                    self.size_distribution_models[model_comps[0]]["_".join(model_comps[2:-1])][
                        model_name] = pickle.load(dmf)

        track_model_files = sorted(glob(model_path + "*_track.pkl"))
        if len(track_model_files) > 0:
            for track_model_file in track_model_files:
                model_comps = track_model_file.split("/")[-1][:-4].split("_")
                group = model_comps[0]
                model_name = model_comps[1].replace("-", " ")
                model_type = model_comps[2]
                if model_type not in self.track_models.keys():
                    self.track_models[model_type] = {}
                if group not in self.track_models[model_type].keys():
                    self.track_models[model_type][group] = {}
                with open(track_model_file, "rb") as tmf:
                    self.track_models[model_type][group][model_name] = pickle.load(tmf)

    def output_forecasts_json(self, forecasts,
                              condition_model_names,
                              size_model_names,
                              dist_model_names,
                              track_model_names,
                              json_data_path,
                              out_path):
        """
        Output forecast values to geoJSON file format.
        :param forecasts:
        :param condition_model_names:
        :param size_model_names:
        :param track_model_names:
        :param json_data_path:
        :param out_path:
        :return:
        """
        total_tracks = self.data["forecast"]["total"]
        for r in np.arange(total_tracks.shape[0]):
            track_id = total_tracks.loc[r, "Track_ID"]
            print(track_id)
            track_num = track_id.split("_")[-1]
            ensemble_name = total_tracks.loc[r, "Ensemble_Name"]
            member = total_tracks.loc[r, "Ensemble_Member"]
            group = self.data["forecast"]["member"].loc[self.data["forecast"]["member"]["Ensemble_Member"] == member,
                                                        self.group_col].values[0]
            run_date = track_id.split("_")[-4][:8]
            step_forecasts = {}
            for ml_model in condition_model_names:
                step_forecasts["condition_" + ml_model.replace(" ", "-")] = forecasts["condition"][group].loc[
                    forecasts["condition"][group]["Track_ID"] == track_id, ml_model]
            for ml_model in size_model_names:
                step_forecasts["size_" + ml_model.replace(" ", "-")] = forecasts["size"][group][ml_model].loc[
                    forecasts["size"][group][ml_model]["Track_ID"] == track_id]
            for ml_model in dist_model_names:
                step_forecasts["dist_" + ml_model.replace(" ", "-")] = forecasts["dist"][group][ml_model].loc[
                    forecasts["dist"][group][ml_model]["Track_ID"] == track_id]
            for model_type in forecasts["track"].keys():
                for ml_model in track_model_names:
                    mframe = forecasts["track"][model_type][group][ml_model]
                    step_forecasts[model_type + "_" + ml_model.replace(" ", "-")] = mframe.loc[
                        mframe["Track_ID"] == track_id]
            json_file_name = "{0}_{1}_{2}_model_track_{3}.json".format(ensemble_name,
                                                                       run_date,
                                                                       member,
                                                                       track_num)
            full_json_path = json_data_path + "/".join([run_date, member]) + "/" + json_file_name
            with open(full_json_path) as json_file_obj:
                try:
                    track_obj = json.load(json_file_obj)
                except FileNotFoundError:
                    print(full_json_path + " not found")
                    continue
            for f, feature in enumerate(track_obj['features']):
                del feature['properties']['attributes']
                for model_name, fdata in step_forecasts.items():
                    ml_model_name = model_name.split("_")[1]
                    if "condition" in model_name:
                        feature['properties'][model_name] = fdata.values[f]
                    else:
                        predcols = []
                        for col in fdata.columns:
                            if ml_model_name in col:
                                predcols.append(col)
                        feature['properties'][model_name] = fdata.loc[:, predcols].values[f].tolist()
            full_path = []
            for part in [run_date, member]:
                full_path.append(part)
                if not os.access(out_path + "/".join(full_path), os.R_OK):
                    try:
                        os.mkdir(out_path + "/".join(full_path))
                    except OSError:
                        print("directory already created")
            out_json_filename = out_path + "/".join(full_path) + "/" + json_file_name
            with open(out_json_filename, "w") as out_json_obj:
                json.dump(track_obj, out_json_obj, indent=1, sort_keys=True)
        return

    def output_forecasts_csv(self, forecasts, mode, csv_path, run_date_format="%Y%m%d-%H%M"):
        """
        Output hail forecast values to csv files by run date and ensemble member.

        Args:
            forecasts:
            mode:
            csv_path:
        Returns:
        """
        merged_forecasts = pd.merge(forecasts["condition"],
                                    forecasts["dist"],
                                    on=["Step_ID","Track_ID","Ensemble_Member","Forecast_Hour"])
        all_members = self.data[mode]["combo"]["Ensemble_Member"]
        members = np.unique(all_members)
        all_run_dates = pd.DatetimeIndex(self.data[mode]["combo"]["Run_Date"])
        run_dates = pd.DatetimeIndex(np.unique(all_run_dates))
        print(run_dates)
        for member in members:
            for run_date in run_dates:
                mem_run_index = (all_run_dates == run_date) & (all_members == member)
                member_forecast = merged_forecasts.loc[mem_run_index]
                member_forecast.to_csv(join(csv_path, "hail_forecasts_{0}_{1}_{2}.csv".format(self.ensemble_name,
                                                                                              member,
                                                                                              run_date.strftime
																							  (run_date_format))))
        return

    def output_forecasts_json_parallel(self, forecasts,
                                       condition_model_names,
                                       size_model_names,
                                       dist_model_names,
                                       track_model_names,
                                       json_data_path,
                                       out_path,
                                       num_procs):
        pool = Pool(num_procs)
        total_tracks = self.data["forecast"]["total_group"]
        for r in total_tracks.index:
            track_id = "{0}".format(total_tracks.loc[r, "Track_ID"])
            print(track_id)
            track_num = track_id.split("_")[-1]
            ensemble_name = "{0}".format(total_tracks.loc[r, "Ensemble_Name"])
            member = "{0}".format(total_tracks.loc[r, "Ensemble_Member"])
            group = total_tracks.loc[r, self.group_col]
            run_date = track_id.split("_")[-4][:8]
            step_forecasts = {}
            for ml_model in condition_model_names:
                step_forecasts["condition_" + ml_model.replace(" ", "-")] = forecasts["condition"][group].loc[
                    forecasts["condition"][group]["Track_ID"] == track_id, ml_model].copy()
            # for ml_model in size_model_names:
            #    step_forecasts["size_" + ml_model.replace(" ", "-")] = forecasts["size"][group][ml_model].loc[
            #        forecasts["size"][group][ml_model]["Track_ID"] == track_id].copy()
            for ml_model in dist_model_names:
                step_forecasts["dist_" + ml_model.replace(" ", "-")] = forecasts["dist"][group][ml_model].loc[
                    forecasts["dist"][group][ml_model]["Track_ID"] == track_id].copy()
            # for model_type in forecasts["track"].keys():
            #    for ml_model in track_model_names:
            #        mframe = forecasts["track"][model_type][group][ml_model]
            #        step_forecasts[model_type + "_" + ml_model.replace(" ", "-")] = mframe.loc[
            #            mframe["Track_ID"] == track_id].copy()
            pool.apply_async(output_forecast, (step_forecasts, run_date, ensemble_name, member, track_num,
                                               json_data_path,
                                               out_path))
        pool.close()
        pool.join()
        return


def output_forecast(step_forecasts, run_date, ensemble_name, member, track_num, json_data_path, out_path):
    json_file_name = "{0}_{1}_{2}_model_track_{3}.json".format(ensemble_name,
                                                               run_date,
                                                               member,
                                                               track_num)
    full_json_path = json_data_path + "/".join([run_date, member]) + "/" + json_file_name
    try:
        json_file_obj = open(full_json_path)
        track_obj = json.load(json_file_obj)
        json_file_obj.close()
    except IOError:
        print(full_json_path + " not found")
        return
    for f, feature in enumerate(track_obj['features']):
        del feature['properties']['attributes']
        for model_name, fdata in step_forecasts.items():
            ml_model_name = model_name.split("_")[1]
            if "condition" in model_name:
                feature['properties'][model_name] = fdata.values[f]
            else:
                predcols = []
                for col in fdata.columns:
                    if ml_model_name in col:
                        predcols.append(col)
                feature['properties'][model_name] = fdata.loc[:, predcols].values[f].tolist()
    full_path = []
    for part in [run_date, member]:
        full_path.append(part)
        if not os.access(out_path + "/".join(full_path), os.R_OK):
            try:
                os.mkdir(out_path + "/".join(full_path))
            except OSError:
                print("directory already created")
    out_json_filename = out_path + "/".join(full_path) + "/" + json_file_name
    try:
        out_json_obj = open(out_json_filename, "w")
        json.dump(track_obj, out_json_obj, indent=1, sort_keys=True)
        out_json_obj.close()
    except IOError:
        print(out_json_filename + " not found")
        return
    return



import numpy as np
import pandas as pd
import json
from glob import glob
from hagelslag.evaluation.ProbabilityMetrics import DistributedCRPS, DistributedReliability, DistributedROC
from scipy.stats import gamma


class ObjectEvaluator(object):
    """
    ObjectEvaluator performs a statistical evaluation of object-based severe weather forecasts.

    ObjectEvaluator loads forecast and observation files for a particular ensemble member and model run and then matches
    the forecasts with their assigned observations. Verification statistics can be calculated on the full dataset
    or on subsets selected based on filter queries.

    Attributes:
        run_date (datetime.datetime): The date marking the start of the model run.
        ensemble_name (str): The name of the ensemble or NWP model being used.
        ensemble_member (str): The name of the ensemble member being evaluated.
        model_names (list): The names of the machine learning models being evaluated
        model_types (list): The types of machine learning models being evaluated.
        forecast_bins (dict of str and numpy.ndarray pairs): For machine learning models forecasting a discrete pdf,
            this specifies the bin labels used.
        dist_thresholds (array): Thresholds used to discretize probability distribution forecasts.
        forecast_json_path (str): Full path to the directory containing all json files with the forecast values.
        track_data_csv_path (str): Full path to the directory containing the csv data files used for training.
        metadata_columns (list): Columns pulled from track data csv files.
        type_cols (dict): Map between forecast type used in json files and observation column in csv files
        forecasts (dict): Dictionary of DataFrames containing forecast information from csv files
        matched_forecasts (dict): Forecasts merged with observation information.

    """
    def __init__(self, run_date, ensemble_name, ensemble_member, model_names, model_types, forecast_bins,
                 dist_thresholds, forecast_json_path, track_data_csv_path):
        self.run_date = run_date
        self.ensemble_name = ensemble_name
        self.ensemble_member = ensemble_member
        self.model_names = model_names
        self.model_types = model_types
        self.forecast_bins = forecast_bins
        self.dist_thresholds = dist_thresholds
        self.forecast_json_path = forecast_json_path
        self.track_data_csv_path = track_data_csv_path
        self.metadata_columns = ["Track_ID", "Obs_Track_ID", "Ensemble_Name", "Ensemble_Member", "Forecast_Hour",
                                 "Step_Duration", "Total_Duration", "Area"]
        self.type_cols = {"size": "Hail_Size",
                          "translation-x": "Translation_Error_X",
                          "translation-y": "Translation_Error_Y",
                          "start-time": "Start_Time_Error",
                          "dist": ["Shape", "Location", "Scale"],
                          "condition": "Hail_Size"}
        self.forecasts = {}
        self.obs = None
        self.matched_forecasts = {}
        for model_type in self.model_types:
            self.forecasts[model_type] = {}
            for model_name in self.model_names[model_type]:
                self.forecasts[model_type][model_name] = pd.DataFrame(columns=self.metadata_columns +
                                                                      list(self.forecast_bins[model_type].astype(str)))

    def load_forecasts(self):
        """
        Loads the forecast files and gathers the forecast information into pandas DataFrames.
        """
        forecast_path = self.forecast_json_path + "/{0}/{1}/".format(self.run_date.strftime("%Y%m%d"),
                                                                     self.ensemble_member)
        forecast_files = sorted(glob(forecast_path + "*.json"))
        for forecast_file in forecast_files:
            file_obj = open(forecast_file)
            json_obj = json.load(file_obj)
            file_obj.close()
            track_id = json_obj['properties']["id"]
            obs_track_id = json_obj['properties']["obs_track_id"]
            forecast_hours = json_obj['properties']['times']
            duration = json_obj['properties']['duration']
            for f, feature in enumerate(json_obj['features']):
                area = np.sum(feature["properties"]["masks"])
                step_id = track_id + "_{0:02d}".format(f)
                for model_type in self.model_types:
                    for model_name in self.model_names[model_type]:
                        prediction = feature['properties'][model_type + "_" + model_name.replace(" ", "-")]
                        if model_type == "condition":
                            prediction = [prediction]
                        row = [track_id, obs_track_id, self.ensemble_name, self.ensemble_member, forecast_hours[f],
                               f + 1, duration, area] + prediction
                        self.forecasts[model_type][model_name].loc[step_id] = row

    def load_obs(self):
        """
        Loads the track total and step files and merges the information into a single data frame.
        """
        track_total_file = self.track_data_csv_path + \
            "track_total_{0}_{1}_{2}.csv".format(self.ensemble_name,
                                                 self.ensemble_member,
                                                 self.run_date.strftime("%Y%m%d"))
        track_step_file = self.track_data_csv_path + \
            "track_step_{0}_{1}_{2}.csv".format(self.ensemble_name,
                                                self.ensemble_member,
                                                self.run_date.strftime("%Y%m%d"))
        track_total_cols = ["Track_ID", "Translation_Error_X", "Translation_Error_Y", "Start_Time_Error"]
        track_step_cols = ["Step_ID", "Track_ID", "Hail_Size", "Shape", "Location", "Scale"]
        track_total_data = pd.read_csv(track_total_file, usecols=track_total_cols)
        track_step_data = pd.read_csv(track_step_file, usecols=track_step_cols)
        obs_data = pd.merge(track_step_data, track_total_data, on="Track_ID", how="left")
        self.obs = obs_data

    def merge_obs(self):
        """
        Match forecasts and observations.
        """
        for model_type in self.model_types:
            self.matched_forecasts[model_type] = {}
            for model_name in self.model_names[model_type]:
                self.matched_forecasts[model_type][model_name] = pd.merge(self.forecasts[model_type][model_name],
                                                                          self.obs, right_on="Step_ID", how="left",
                                                                          left_index=True)

    def crps(self, model_type, model_name, condition_model_name, condition_threshold, query=None):
        """
        Calculates the cumulative ranked probability score (CRPS) on the forecast data.

        Args:
            model_type: model type being evaluated.
            model_name: machine learning model being evaluated.
            condition_model_name: Name of the hail/no-hail model being evaluated
            condition_threshold: Threshold for using hail size CDF
            query: pandas query string to filter the forecasts based on the metadata


        Returns:
            a DistributedCRPS object
        """

        def gamma_cdf(x, a, loc, b):
            if a == 0 or b == 0:
                cdf = np.ones(x.shape)
            else:
                cdf = gamma.cdf(x, a, loc, b)
            return cdf

        crps_obj = DistributedCRPS(self.dist_thresholds)
        if query is not None:
            sub_forecasts = self.matched_forecasts[model_type][model_name].query(query)
            sub_forecasts = sub_forecasts.reset_index(drop=True)
            condition_forecasts = self.matched_forecasts["condition"][condition_model_name].query(query)
            condition_forecasts = condition_forecasts.reset_index(drop=True)
        else:
            sub_forecasts = self.matched_forecasts[model_type][model_name]
            condition_forecasts = self.matched_forecasts["condition"][condition_model_name]
        if sub_forecasts.shape[0] > 0:
            if model_type == "dist":
                forecast_cdfs = np.zeros((sub_forecasts.shape[0], self.dist_thresholds.size))
                for f in range(sub_forecasts.shape[0]):
                    condition_prob = condition_forecasts.loc[f, self.forecast_bins["condition"][0]]
                    if condition_prob >= condition_threshold:
                        f_params = [0, 0, 0]
                    else:
                        f_params = sub_forecasts[self.forecast_bins[model_type]].values[f]
                    forecast_cdfs[f] = gamma_cdf(self.dist_thresholds, f_params[0], f_params[1], f_params[2])
                obs_cdfs = np.array([gamma_cdf(self.dist_thresholds, *params)
                                    for params in sub_forecasts[self.type_cols[model_type]].values])
                crps_obj.update(forecast_cdfs, obs_cdfs)
            else:
                crps_obj.update(sub_forecasts[self.forecast_bins[model_type].astype(str)].values,
                                sub_forecasts[self.type_cols[model_type]].values)

        return crps_obj

    def roc(self, model_type, model_name, intensity_threshold, prob_thresholds, query=None):
        """
        Calculates a ROC curve at a specified intensity threshold.

        Args:
            model_type: type of model being evaluated (e.g. size).
            model_name: machine learning model being evaluated
            intensity_threshold: forecast bin used as the split point for evaluation
            prob_thresholds: Array of probability thresholds being evaluated.
            query: str to filter forecasts based on values of forecasts, obs, and metadata.

        Returns:
             A DistributedROC object
        """
        roc_obj = DistributedROC(prob_thresholds, 0.5)
        if query is not None:
            sub_forecasts = self.matched_forecasts[model_type][model_name].query(query)
            sub_forecasts = sub_forecasts.reset_index(drop=True)
        else:
            sub_forecasts = self.matched_forecasts[model_type][model_name]
        obs_values = np.zeros(sub_forecasts.shape[0])
        if sub_forecasts.shape[0] > 0:
            if model_type == "dist":
                forecast_values = np.array([gamma_sf(intensity_threshold, *params)
                                            for params in sub_forecasts[self.forecast_bins[model_type]].values])
                obs_probs = np.array([gamma_sf(intensity_threshold, *params)
                                    for params in sub_forecasts[self.type_cols[model_type]].values])
                obs_values[obs_probs >= 0.01] = 1
            elif len(self.forecast_bins[model_type]) > 1:
                fbin = np.argmin(np.abs(self.forecast_bins[model_type] - intensity_threshold))
                forecast_values = 1 - sub_forecasts[self.forecast_bins[model_type].astype(str)].values.cumsum(axis=1)[:,
                                    fbin]
                obs_values[sub_forecasts[self.type_cols[model_type]].values >= intensity_threshold] = 1
            else:
                forecast_values = sub_forecasts[self.forecast_bins[model_type].astype(str)[0]].values
                obs_values[sub_forecasts[self.type_cols[model_type]].values >= intensity_threshold] = 1
            roc_obj.update(forecast_values, obs_values)
        return roc_obj

    def reliability(self, model_type, model_name, intensity_threshold, prob_thresholds, query=None):
        """
        Calculate reliability statistics based on the probability of exceeding a specified threshold.

        Args:
            model_type: type of model being evaluated.
            model_name: Name of the machine learning model being evaluated.
            intensity_threshold: forecast bin used as the split point for evaluation.
            prob_thresholds: Array of probability thresholds being evaluated.
            query: str to filter forecasts based on values of forecasts, obs, and metadata.

        Returns:
            A DistributedReliability object.
        """
        rel_obj = DistributedReliability(prob_thresholds, 0.5)
        if query is not None:
            sub_forecasts = self.matched_forecasts[model_type][model_name].query(query)
            sub_forecasts = sub_forecasts.reset_index(drop=True)
        else:
            sub_forecasts = self.matched_forecasts[model_type][model_name]
        obs_values = np.zeros(sub_forecasts.shape[0])
        if sub_forecasts.shape[0] > 0:
            if model_type == "dist":
                forecast_values = np.array([gamma_sf(intensity_threshold, *params)
                                            for params in sub_forecasts[self.forecast_bins[model_type]].values])
                obs_probs = np.array([gamma_sf(intensity_threshold, *params)
                                    for params in sub_forecasts[self.type_cols[model_type]].values])
                obs_values[obs_probs >= 0.01] = 1
            elif len(self.forecast_bins[model_type]) > 1:
                fbin = np.argmin(np.abs(self.forecast_bins[model_type] - intensity_threshold))
                forecast_values = 1 - sub_forecasts[self.forecast_bins[model_type].astype(str)].values.cumsum(axis=1)[:,
                                    fbin]
                obs_values[sub_forecasts[self.type_cols[model_type]].values >= intensity_threshold] = 1
            else:
                forecast_values = sub_forecasts[self.forecast_bins[model_type].astype(str)[0]].values
                obs_values[sub_forecasts[self.type_cols[model_type]].values >= intensity_threshold] = 1
            rel_obj.update(forecast_values, obs_values)
        return rel_obj

    def sample_forecast_max_hail(self, dist_model_name, condition_model_name,
                                 num_samples, condition_threshold=0.5, query=None):
        """
        Samples every forecast hail object and returns an empirical distribution of possible maximum hail sizes.

        Hail sizes are sampled from each predicted gamma distribution. The total number of samples equals
        num_samples * area of the hail object. To get the maximum hail size for each realization, the maximum
        value within each area sample is used.

        Args:
            dist_model_name: Name of the distribution machine learning model being evaluated
            condition_model_name: Name of the hail/no-hail model being evaluated
            num_samples: Number of maximum hail samples to draw
            condition_threshold: Threshold for drawing hail samples
            query: A str that selects a subset of the data for evaluation

        Returns:
            A numpy array containing maximum hail samples for each forecast object.
        """
        if query is not None:
            dist_forecasts = self.matched_forecasts["dist"][dist_model_name].query(query)
            dist_forecasts = dist_forecasts.reset_index(drop=True)
            condition_forecasts = self.matched_forecasts["condition"][condition_model_name].query(query)
            condition_forecasts = condition_forecasts.reset_index(drop=True)
        else:
            dist_forecasts = self.matched_forecasts["dist"][dist_model_name]
            condition_forecasts = self.matched_forecasts["condition"][condition_model_name]
        max_hail_samples = np.zeros((dist_forecasts.shape[0], num_samples))
        areas = dist_forecasts["Area"].values
        for f in np.arange(dist_forecasts.shape[0]):
            condition_prob = condition_forecasts.loc[f, self.forecast_bins["condition"][0]]
            if condition_prob >= condition_threshold:
                max_hail_samples[f] = np.sort(gamma.rvs(*dist_forecasts.loc[f, self.forecast_bins["dist"]].values,
                                                        size=(num_samples, areas[f])).max(axis=1))
        return max_hail_samples

    def sample_obs_max_hail(self, dist_model_name, num_samples, query=None):
        if query is not None:
            dist_obs = self.matched_forecasts["dist"][dist_model_name].query(query)
            dist_obs = dist_obs.reset_index(drop=True)
        else:
            dist_obs = self.matched_forecasts["dist"][dist_model_name]
        max_hail_samples = np.zeros((dist_obs.shape[0], num_samples))
        areas = dist_obs["Area"].values
        for f in np.arange(dist_obs.shape[0]):
            dist_params = dist_obs.loc[f, self.type_cols["dist"]].values
            if dist_params[0] > 0:
                max_hail_samples[f] = np.sort(gamma.rvs(*dist_params,
                                                        size=(num_samples, areas[f])).max(axis=1))
        return max_hail_samples

    def max_hail_sample_crps(self, forecast_max_hail, obs_max_hail):
        crps = DistributedCRPS(thresholds=self.dist_thresholds)
        if forecast_max_hail.shape[0] > 0:
            forecast_cdfs = np.array([np.searchsorted(fs, self.dist_thresholds, side="right")
                                      for fs in forecast_max_hail]) / float(forecast_max_hail.shape[1])
            obs_cdfs = np.array([np.searchsorted(obs, self.dist_thresholds, side="right")
                                 for obs in obs_max_hail]) / float(obs_max_hail.shape[1])
            crps.update(forecast_cdfs, obs_cdfs)
        return crps




def gamma_sf(x, a, loc, b):
    if a == 0 or b == 0:
        sf = 0
    else:
        sf = gamma.sf(x, a, loc, b)
    return sf
